from Orange.data import StringVariable, Table
from Orange.widgets import gui
from Orange.widgets.settings import DomainContextHandler, ContextSetting
from Orange.widgets.utils.itemmodels import DomainModel
from Orange.widgets.widget import Input, Msg

from Orange.widgets.data.owsave import OWSaveBase
from orangecontrib.network.network import readwrite
from orangecontrib.network.network.base import Network
from orangecontrib.network.network.readwrite import PajekReader
from orangewidget.utils.widgetpreview import WidgetPreview


class OWNxSave(OWSaveBase):
    name = "Save Network"
    description = "Save network to an output file."
    icon = "icons/NetworkSave.svg"

    writers = [PajekReader]
    filters = {f"{w.DESCRIPTION} (*{w.EXTENSIONS[0]})": w
               for w in writers}

    class Inputs:
        network = Input("Network", Network, default=True)

    class Error(OWSaveBase.Error):
        multiple_edge_types = Msg("Can't save network with multiple edge types")

    settingsHandler = DomainContextHandler()
    label_variable = ContextSetting(None)

    def __init__(self):
        super().__init__(2)

        self.label_model = DomainModel(
            placeholder="(None)", valid_types=(StringVariable, ))
        box = gui.hBox(None)
        gui.widgetLabel(box, "Node label: ")
        gui.comboBox(
            box, self, "label_variable",
            tooltip="Choose the variables that will be used as a label",
            model=self.label_model),

        self.grid.addWidget(box, 0, 0, 1, 2)
        self.grid.setRowMinimumHeight(1, 8)
        self.adjustSize()

    @Inputs.network
    def set_network(self, network):
        self.closeContext()

        self.data = network
        if network is None:
            return
        if len(network.edges) > 1:
            self.Error.multiple_edge_types()
            self.data = None
            return
        self.Error.multiple_edge_types.clear()

        if isinstance(network.nodes, Table):
            self.controls.label_variable.setEnabled(True)
            domain = network.nodes.domain
            self.label_model.set_domain(domain)
            for attr in domain.metas:
                if attr.name == "node_label":
                    self.label_variable = attr
                    break
            self.openContext(domain)
        else:
            self.label_model.set_domain(None)
            self.label_variable = None
            self.controls.label_variable.setEnabled(False)

        self.on_new_input()

    def save_file(self):
        if not self.filename:
            self.save_file_as()
            return

        self.Error.general_error.clear()
        if self.data is None or not self.filename or self.writer is None:
            return
        try:
            net = self.data
            if self.label_variable is not None:
                labels = net.nodes.get_column(self.label_variable)
            else:
                labels = range(1, net.number_of_nodes() + 1)
            self.writer.write(self.filename, net, labels)
        except IOError as err_value:
            self.Error.general_error(str(err_value))

    def send_report(self):
        self.report_items((
            ("Node labels",
             self.label_variable.name if self.label_variable else "none"),
            ("File name", self.filename or "not set"),
        ))


def main_with_annotation():
    from AnyQt.QtWidgets import QApplication
    from OWNxFile import OWNxFile
    app = QApplication([])
    file_widget = OWNxFile()
    file_widget.Outputs.network.send = WidgetPreview(OWNxSave).run
    file_widget.open_net_file("../networks/leu_by_genesets.net")


def main_without_annotation():
    net = readwrite.read_pajek("../networks/leu_by_genesets.net")
    WidgetPreview(OWNxSave).run(net)


if __name__ == "__main__":  # pragma: no cover
    main_with_annotation()



import unittest
from unittest.mock import Mock

import numpy as np
from AnyQt.QtCore import QLineF
from AnyQt.QtGui import QPen, QBrush, QColor

from Orange.widgets.tests.base import GuiTest
from orangecontrib.network.widgets.graphview import PlotVarWidthCurveItem

class TestPlotVarWidthCurveItem(GuiTest):
    def line_to_tuple(self, line: QLineF):
        return line.x1(), line.y1(), line.x2(), line.y2()

    def test_coordinates(self):
        curve = PlotVarWidthCurveItem(
            directed=False,
            x=np.array([0, 18, 2, 2]), y=np.array([10, 10, 20, 40]),
            size=np.ones(4))
        painter = Mock()
        painter.worldTransform = Mock()
        painter.worldTransform.return_value.m11 = Mock(return_value=1)
        painter.worldTransform.return_value.m22 = Mock(return_value=1)

        curve.pen.width = lambda: 3
        curve.paint(painter, None, None)
        # coordinates are moved by sizes + width / 3 = 1 + 1 = 5
        self.assertEqual(
            [self.line_to_tuple(call[0][0])
             for call in painter.drawLine.call_args_list],
            [(2, 10, 16, 10), (2, 22, 2, 38)])

        painter.reset_mock()
        curve.pen.width = lambda: 6
        curve.setData(
            x=np.array([2, 2]), y=np.array([3, 20]),
            size=np.ones(2))
        curve.paint(painter, None, None)
        self.assertEqual(
            [self.line_to_tuple(call[0][0])
             for call in painter.drawLine.call_args_list],
            [(2, 6, 2, 17)])

        painter.reset_mock()
        s = np.arange(100) / 120
        curve.setData(
            np.arange(100, dtype=float), np.arange(100, 200, dtype=float),
            size=s)
        curve.paint(painter, None, None)
        s2 = np.sqrt(2) / 2
        exp = np.array([self.line_to_tuple(call[0][0])
                        for call in painter.drawLine.call_args_list])
        act = np.array([(2 * x + (2 + s[2 * x]) * s2,
                         2 * x + 100 + (2 + s[2 * x]) * s2,
                         2 * x + 1 - (2 + s[2 * x + 1]) * s2,
                         2 * x + 101 - (2 + s[2 * x + 1]) * s2) for x in range(50)])
        np.testing.assert_almost_equal(exp, act)

        painter.reset_mock()
        curve.setData(None, None)
        curve.paint(painter, None, None)
        painter.drawline.assert_not_called()

    def test_widths(self, ):
        def draw_line(*_):
            pens.append(painter.setPen.call_args[0][0].width())

        painter = Mock()
        painter.drawLine = draw_line
        painter.worldTransform = Mock()
        painter.worldTransform.return_value.m11 = Mock(return_value=1)
        painter.worldTransform.return_value.m22 = Mock(return_value=1)

        curve = PlotVarWidthCurveItem(
            directed=False,
            x=np.arange(4, dtype=float), y=np.arange(10, 14, dtype=float),
            size=np.ones(4), widths=np.array([2, 3]))

        pens = []
        curve.paint(painter, None, None)
        self.assertEqual(pens, [2, 3])

        curve.setWidths(np.array([10, 11]))
        pens = []
        curve.paint(painter, None, None)
        self.assertEqual(pens, [10, 11])

        curve.setData(
            np.arange(100, dtype=float), np.arange(100, 200, dtype=float),
            widths=np.arange(1000, 1050), size=np.ones(100))
        pens = []
        curve.paint(painter, None, None)
        self.assertEqual(pens, list(range(1000, 1050)))

        curve.setWidths(None)
        curve.setPen(QPen(QBrush(QColor(0, 0, 0)), 42))
        pens = []
        curve.paint(painter, None, None)
        self.assertEqual(pens, [42] * 50)

        curve.setData(
            np.arange(10, dtype=float), np.arange(100, 110, dtype=float),
            pen=QPen(QBrush(QColor(0, 0, 0)), 10), size=np.ones(10))
        pens = []
        curve.paint(painter, None, None)
        self.assertEqual(pens, [10] * 5)


if __name__ == "__main__":
    unittest.main()



from collections import namedtuple

import numpy as np
import scipy.sparse as sp

from orangecontrib.network import Network

NoWeights, WeightConnections, WeightWeightedConnections,\
    WeightGeo, WeightGeoDeg, WeightInput, WeightOutput, WeightMin, WeightMax \
    = range(9)


def to_single_mode(net, mode_mask, conn_mask, weighting):
    """
    Convert two-mode network into a single mode

    Args:
        net: network to convert
        mode_mask (boolean array): a mask with nodes to connect
        conn_mask (boolean array): a mask with nodes to use for connecting
        weighting (int): normalization for edge weigthts

    Returns:
        single-mode network
    """
    mode_edges = _filtered_edges(net, mode_mask, conn_mask)
    new_edges = Weighting[weighting].func(mode_edges, np.sum(mode_mask))
    return Network(net.nodes[mode_mask], new_edges)


def _normalize(a, *nominators):
    for nominator in nominators:
        with np.errstate(divide='ignore', invalid='ignore'):
            inv = np.reciprocal(nominator)
        inv[nominator == 0] = 1
        a = a.multiply(inv)
    return a


def _dot_edges(normalization):
    def norm_dot(edges, dim):
        edges = normalization(
            edges, wuu=lambda: edges.sum(axis=0), wvv=lambda: edges.sum(axis=1))
        new_edges = np.dot(edges, edges.T).tocoo()
        mask = np.logical_and(
            new_edges.row < new_edges.col,  new_edges.data != 0)

        return sp.csr_matrix(
            (new_edges.data[mask], (new_edges.row[mask], new_edges.col[mask])),
            shape=(dim, dim))
    return norm_dot


def _weight_no_weights(edges, dim):
    new_edges = _weight_connections(edges, dim)
    new_edges[new_edges.nonzero()] = 1
    return new_edges


@_dot_edges
def _weight_connections(edges, **_):
    edges = edges.copy()
    edges.data[:] = 1
    return edges


@_dot_edges
def _weight_weighted_connections(edges, **_):
    return edges


@_dot_edges
def _weight_geo(edges, wuu, wvv):
    return _normalize(edges, np.sqrt(wuu()), np.sqrt(wvv()))


@_dot_edges
def _weight_geodeg(edges, **_):
    return _normalize(
        edges,
        np.sqrt(edges.getnnz(axis=0)),
        np.sqrt(edges.getnnz(axis=1)).reshape(edges.shape[1], 1))


@_dot_edges
def _weight_input(edges, wvv, **_):
    return _normalize(edges, wvv())


@_dot_edges
def _weight_output(edges, wuu, **_):
    return _normalize(edges, wuu())


def _norm_min_max(edges, wuu, wvv, f):
    assert isinstance(edges, sp.coo_matrix)
    wuu, wvv = wuu(), wvv()
    edges = edges.copy()
    us, vs = edges.col, edges.row
    edges.data /= f(wuu.A.flatten()[us], wvv.A.flatten()[vs])
    return edges


@_dot_edges
def _weight_min(edges, wuu, wvv):
    return _norm_min_max(edges, wuu, wvv, np.minimum)


@_dot_edges
def _weight_max(edges, wuu, wvv):
    return _norm_min_max(edges, wuu, wvv, np.maximum)


WeightType = namedtuple("WeightType", ["name", "func"])

Weighting = [WeightType(*x) for x in (
    ("No weights", _weight_no_weights),
    ("Number of connections", _weight_connections),
    ("Weighted number of connections", _weight_weighted_connections),
    ("Geometric normalization", _weight_geo),
    ("Geometric normalization by degrees", _weight_geodeg),
    ("Normalization by sum of input weights", _weight_input),
    ("Normalization by sum of output weights", _weight_output),
    ("Normalization by minimal sum of weights", _weight_min),
    ("Normalization by maximal sum of weights", _weight_max)
)]


def _filtered_edges(network, mode_mask, conn_mask):
    """
    Compute a coo_matrix with network edges between nodes in mode_mask and
    conn_mask
    """
    if not network.number_of_edges():
        return sp.coo_matrix((0, 0))
    edges = network.edges[0].edges.tocoo()
    node1, node2, weights = edges.row, edges.col, edges.data
    fwd = mode_mask[node1] * conn_mask[node2]
    back = mode_mask[node2] * conn_mask[node1]
    us = np.hstack((node1[fwd], node2[back]))
    vs = np.hstack((node2[fwd], node1[back]))
    if not len(us):
        return sp.coo_matrix((0, 0))
    data = np.hstack((weights[fwd], weights[back]))
    new_indices = np.cumsum(mode_mask) - 1
    return sp.coo_matrix((data, (new_indices[us], vs)))



import numpy as np
import gensim
from gensim.models import Word2Vec
from Orange.data import ContinuousVariable, Table, Domain

from orangecontrib.network.network.base import DirectedEdges


def alias_preprocess(probas_dict):
    """
        Preprocessing step for alias sampling, which enables drawing samples in O(1).
        Returns triple of equal-length lists: [0] contains candidates, [1] contains the probability
        of selecting the candidates after a biased coin flip, [2] contains reference to the alternative choice
        after the flip.

        Arguments
        ---------
        probas_dict: dict
            Keys are nodes/edges, values are their probabilities

        Useful reference: http://www.keithschwarz.com/darts-dice-coins/ (see Vose's Alias Method)
    """
    num_items = len(probas_dict)
    alias, prob = [None] * num_items, [None] * num_items
    # hold unnormalized probabilities < and >= 1, respectively
    small, large = [], []

    items = []
    working_probas = []
    for i, (curr_item, proba) in enumerate(probas_dict.items()):
        items.append(curr_item)
        unnorm_proba = proba * num_items
        working_probas.append(unnorm_proba)
        if unnorm_proba < 1.0:
            small.append((i, unnorm_proba))
        else:
            large.append((i, unnorm_proba))

    while small and large:
        curr_lower, p_lower = small.pop()
        curr_greater, p_greater = large.pop()

        prob[curr_lower] = p_lower
        alias[curr_lower] = curr_greater

        remaining_greater = (p_greater + p_lower) - 1.0
        if remaining_greater < 1.0:
            small.append((curr_greater, remaining_greater))
        else:
            large.append((curr_greater, remaining_greater))

    while large:
        curr_greater, p_greater = large.pop()
        prob[curr_greater] = 1.0

    # can only be non-empty at this point due to numerical instability
    while small:
        curr_lower, p_lower = small.pop()
        prob[curr_lower] = 1.0

    return items, prob, alias


def alias_draw(candidates, prob, alias):
    """ Single draw of alias sampling - first, perform uniform draw between candidates, then do a biased coin flip
    to determine whether to keep the chosen candidate or take its alternative.

    Arguments
    ---------
    candidates: list
        Raw names of candidates
    prob: list
        Probabilities of choosing items in `candidates` during a biased coin flip
    alias: list
        Reference (indices that index `candidates`) to alternative candidates during a biased coin flip
    """
    bucket_id = np.random.randint(0, len(candidates))
    rand_num = np.random.random()

    if rand_num < prob[bucket_id]:
        return candidates[bucket_id]
    else:
        return candidates[alias[bucket_id]]


def normalize(probas, norm_const):
    # either all weights 0 or some positive, some negative - use softmax
    if norm_const == 0.0:
        normed_probas = {}
        for item, proba in probas.items():
            new_proba = np.exp(proba)
            normed_probas[item] = new_proba
            norm_const += new_proba
    else:
        normed_probas = probas

    return {neigh: w / norm_const for neigh, w in normed_probas.items()}


class Node2Vec:
    def __init__(self, p=0.8, q=0.7, walk_len=80, num_walks=10, emb_size=50,
                 window_size=5, num_epochs=1, prefix=None,
                 callbacks=()):
        self.p = p
        self.q = q
        self.walk_len = walk_len
        self.num_walks = num_walks
        self.emb_size = emb_size
        self.window_size = window_size
        self.num_epochs = num_epochs
        self.callbacks = callbacks

        self._node_probas = {}
        self._edge_probas = {}
        if prefix is not None and isinstance(prefix, str):
            self.feature_prefix = prefix
        else:
            self.feature_prefix = "n2v"

    def __call__(self, G):
        if not G.edges:
            raise ValueError("Network has no edges")

        num_nodes = G.number_of_nodes()
        nodes = np.arange(num_nodes)
        # Node->node probas are needed for the initial step, when there's no previous edge to condition on
        self._node_probas = self.setup_nodes(G, nodes)

        edges_coo = G.edges[0].edges.tocoo(copy=False)
        edges = np.column_stack((edges_coo.row, edges_coo.col))
        self._edge_probas = self.setup_edges(G, edges)

        walks = self._simulate_walks(G)
        walks = [list(map(str, walk)) for walk in walks]

        # gensim changed "size" param to "vector_size" in v. 4.0.0
        # https://github.com/RaRe-Technologies/gensim/wiki/Migrating-from-Gensim-3.x-to-4
        params = dict(window=self.window_size, min_count=0, sg=1, workers=4,
                      callbacks=self.callbacks)
        if gensim.__version__ < "4.0.0":
            params["size"] = self.emb_size
            params["iter"] = self.num_epochs
        else:
            params["vector_size"] = self.emb_size
            params["epochs"] = self.num_epochs
        model = Word2Vec(walks, **params)

        items = G.nodes
        new_attrs = {}
        new_data = np.array([[] for _ in range(num_nodes)])
        class_vars, meta_vars = [], []
        class_data, meta_data = np.array([[] for _ in range(num_nodes)]), np.array([[] for _ in range(num_nodes)])
        if isinstance(items, Table):
            attrs_mask = []
            for attr in items.domain.attributes:
                attrs_mask.append(attr.name not in new_attrs)
                new_attrs[attr.name] = new_attrs.get(attr.name, (len(new_attrs), attr))

            new_data = items.X[:, np.array(attrs_mask, dtype=bool)]
            class_vars, meta_vars = items.domain.class_vars, items.domain.metas
            class_data, meta_data = items.Y, items.metas

        # override existing continuous vars with same names
        for i in range(self.emb_size):
            new_name = "{}_{}".format(self.feature_prefix, i)
            new_attrs[new_name] = (len(new_attrs), ContinuousVariable(new_name))

        new_data = np.hstack((new_data, np.array([model.wv[str(curr_node)] for curr_node in nodes])))
        ordered_attrs = [None] * len(new_attrs)
        for idx, attr in new_attrs.values():
            ordered_attrs[idx] = attr
        new_domain = Domain(ordered_attrs, class_vars, meta_vars)
        new_items = Table(new_domain, new_data, class_data, meta_data)
        return new_items

    def _single_walk(self, start_node):
        # assumes node and edge probabilities have been computed
        curr_walk = [start_node]

        nodes, prob, alias = self._node_probas[start_node]
        # isolated starting node
        if not nodes:
            return curr_walk
        # initial step (no previous node to condition on)
        prev_node, curr_node = start_node, alias_draw(nodes, prob, alias)
        curr_walk.append(curr_node)

        for _ in range(self.walk_len - 2):
            edges, prob, alias = self._edge_probas[(prev_node, curr_node)]
            if len(edges) == 0:
                break

            prev_node, curr_node = alias_draw(edges, prob, alias)
            curr_walk.append(curr_node)

        return curr_walk

    def _simulate_walks(self, G):
        # assumes node and edge probabilities have been computed
        walks = []
        nodes = np.arange(G.number_of_nodes())

        for idx_walk in range(self.num_walks):
            np.random.shuffle(nodes)
            for curr_node in nodes:
                walks.append(self._single_walk(start_node=curr_node))

        return walks

    def setup_nodes(self, G, nodes):
        return {n: alias_preprocess(self.node_probas(G, n)) for n in nodes}

    def node_probas(self, G, from_node):
        probas = {}
        norm_const = 0.0
        edges = G.edges[0]
        edges = edges.edges if isinstance(edges, DirectedEdges) else edges.twoway_edges

        for neigh_node in G.outgoing(from_node):
            edge_w = edges[from_node, neigh_node]
            norm_const += edge_w
            probas[neigh_node] = edge_w

        probas = normalize(probas, norm_const)
        return probas

    def setup_edges(self, G, edges):
        probas = {}
        if isinstance(G.edges[0], DirectedEdges):
            for u, v in edges:
                probas[(u, v)] = alias_preprocess(self.edge_probas(G, u, v))
        else:
            for u, v in edges:
                probas[(u, v)] = alias_preprocess(self.edge_probas(G, u, v))
                probas[(v, u)] = alias_preprocess(self.edge_probas(G, v, u))
        return probas

    def edge_probas(self, G, u, v):
        """ Compute transition probas from `v` to neighbors, taking into account that edge u->v was just traversed. """
        probas = {}
        norm_const = 0.0
        edges = G.edges[0]
        edges = edges.edges if isinstance(edges, DirectedEdges) else edges.twoway_edges

        # for checking if edge exists
        edges_coo = edges.tocoo(copy=False)
        edges_coo = set(zip(edges_coo.row, edges_coo.col))

        for neigh_of_dst in G.outgoing(v):
            edge_w = edges[v, neigh_of_dst]
            if neigh_of_dst == u:
                edge_w = edge_w / self.p
            elif (u, neigh_of_dst) in edges_coo:
                edge_w = edge_w
            else:
                edge_w = edge_w / self.q

            probas[(v, neigh_of_dst)] = edge_w
            norm_const += edge_w

        probas = normalize(probas, norm_const)
        return probas



import unittest
from unittest.mock import patch

import numpy as np

from AnyQt.QtTest import QSignalSpy
from AnyQt.QtWidgets import QLineEdit
from AnyQt.QtCore import QEvent, Qt
from AnyQt.QtGui import QKeyEvent

from orangewidget.tests.base import GuiTest
from Orange.misc import DistMatrix
from Orange.widgets.tests.base import WidgetTest

from orangecontrib.network.widgets.OWNxFromDistances import OWNxFromDistances, \
    Histogram, QIntValidatorWithFixup


class TestOWNxFromDistances(WidgetTest):
    def setUp(self):
        self.widget: OWNxFromDistances = self.create_widget(OWNxFromDistances)

        # Put non-zero elements in the diagonal to check that the widget ignores them
        self.distances = DistMatrix(np.array([[0., 1, 2, 5, 10],
                                              [1, -1, 5, 5, 13],
                                              [2, 5, 2, 6, 13],
                                              [5, 5, 6, 3, 15],
                                              [10, 13, 13, 15, 0]]))

    @staticmethod
    def set_edit(edit, value):
        edit.setText(str(value))
        edit.textChanged.emit(str(value))
        edit.textEdited.emit(str(value))
        edit.returnPressed.emit()
        edit.editingFinished.emit()

    def _assert_controls_enabled(self, enabled):
        widget = self.widget
        self.assertEqual(widget.threshold_edit.isEnabled(), enabled)
        self.assertEqual(widget.edges_edit.isEnabled(), enabled)
        self.assertEqual(widget.density_edit.isEnabled(), enabled)

    def test_set_weird_matrix(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)
        self.assertEqual(widget.eff_distances, 10)
        self.assertIsNotNone(self.get_output(widget.Outputs.network))
        self._assert_controls_enabled(True)

        self.send_signal(widget.Inputs.distances, None)
        self.assertIsNone(self.get_output(widget.Outputs.network))
        self._assert_controls_enabled(False)

        self.send_signal(widget.Inputs.distances, self.distances)
        self.assertIsNotNone(self.get_output(widget.Outputs.network))
        self._assert_controls_enabled(True)

        self.send_signal(widget.Inputs.distances, DistMatrix(np.zeros((0, 0))))
        self.assertIsNone(self.get_output(widget.Outputs.network))
        self._assert_controls_enabled(False)

        self.send_signal(widget.Inputs.distances, DistMatrix(np.array([[1]])))
        self.assertIsNone(self.get_output(widget.Outputs.network))
        self._assert_controls_enabled(False)

        self.send_signal(widget.Inputs.distances, None)
        self.assertIsNone(self.get_output(widget.Outputs.network))
        self._assert_controls_enabled(False)

        self.send_signal(widget.Inputs.distances,
                         DistMatrix(np.array([[0, 1],
                                              [1, 0]])))
        self.assertIsNotNone(self.get_output(widget.Outputs.network))
        self._assert_controls_enabled(True)

    @patch("orangecontrib.network.widgets.OWNxFromDistances.Histogram.set_graph")
    def test_compute_histogram_symmetric(self, set_graph):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)
        np.testing.assert_almost_equal(widget.thresholds, [1, 2, 5, 6, 10, 13, 15])
        np.testing.assert_almost_equal(widget.cumfreqs, [1, 2, 5, 6, 7, 9, 10])
        set_graph.assert_called_with(widget.thresholds, widget.cumfreqs)

        # The matrix is symmetric, so the number of bins is below 1000 ->
        # the histogram is computed exactly
        distances = np.zeros((40, 40))
        for i in range(40):
            distances[:i, i] = distances[i, :i] = np.arange(i)
        distances = DistMatrix(distances)
        self.send_signal(widget.Inputs.distances, distances)
        np.testing.assert_almost_equal(widget.thresholds, np.arange(39))
        set_graph.assert_called_with(widget.thresholds, widget.cumfreqs)

        # Even though the matrix is symmetric, the number of bins is above 1000
        distances = np.zeros((50, 50))
        for i in range(50):
            distances[:i, i] = distances[i, :i] = np.arange(i)
        distances = DistMatrix(distances)
        self.send_signal(widget.Inputs.distances, distances)
        np.testing.assert_almost_equal(
            widget.thresholds[:5], [0.048, 0.096, 0.144, 0.192, 0.24])
        set_graph.assert_called_with(widget.thresholds, widget.cumfreqs)

    @patch("orangecontrib.network.widgets.OWNxFromDistances.Histogram.set_graph")
    def test_compute_histogram_asymmetric(self, set_graph):
        widget = self.widget

        self.distances[0, 1] = 1.5
        self.send_signal(widget.Inputs.distances, self.distances)
        np.testing.assert_almost_equal(widget.thresholds, [1, 1.5, 2, 5, 6, 10, 13, 15])
        np.testing.assert_almost_equal(widget.cumfreqs, [1, 2, 4, 10, 12, 14, 18, 20])
        set_graph.assert_called_with(widget.thresholds, widget.cumfreqs)

        distances = DistMatrix(np.array(np.arange(40 * 40).reshape((40, 40))))
        self.send_signal(widget.Inputs.distances, distances)
        np.testing.assert_almost_equal(
            widget.thresholds[:5], [2.597, 4.194, 5.791, 7.388, 8.985])
        np.testing.assert_almost_equal(
            widget.cumfreqs[:5], [2, 4, 5, 7, 8])
        self.assertEqual(widget.cumfreqs[-1], 40 * 39)
        set_graph.assert_called_with(widget.thresholds, widget.cumfreqs)

    def test_set_symmetric(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)
        self.assertTrue(widget.symmetric)
        self.assertEqual(widget.eff_distances, 10)

        self.distances[0, 1] = 2
        self.send_signal(widget.Inputs.distances, self.distances)
        self.assertFalse(widget.symmetric)
        self.assertEqual(widget.eff_distances, 20)

    def test_set_threshold_from_density_symmetric(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)

        for density, threshold in [(0, 0), (10, 1), (20, 2), (30, 5), (40, 5),
                                   (50, 5), (60, 6), (70, 10), (80, 13),
                                   (90, 13), (100, 15)]:
            self.set_edit(widget.density_edit, density)
            self.assertEqual(widget.threshold, threshold,
                             msg=f"at density={density}")

    def test_set_threshold_from_density_asymmetric(self):
        widget = self.widget
        self.distances[0, 1] = 2
        self.send_signal(widget.Inputs.distances, self.distances)

        for density, threshold in [(0, 0), (10, 2), (20, 2), (21, 5), (50, 5),
                                   (51, 6), (60, 6), (70, 10), (80, 13),
                                   (90, 13), (100, 15)]:
            self.set_edit(widget.density_edit, density)
            self.assertEqual(widget.threshold, threshold,
                             msg=f"at density={density}")

    def test_set_threshold_from_edges(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)

        for edges, threshold in [(0, 0), (1, 1), (2, 2), (3, 5), (4, 5), (5, 5),
                                 (6, 6), (7, 10), (8, 13), (9, 13), (10, 15)]:
            self.set_edit(widget.edges_edit, edges)
            self.assertEqual(widget.threshold, threshold,
                             msg=f"at edges={edges}")

    def test_edges_from_threshold_symmetric(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)

        for threshold, edges in [(0, 0), (1, 1), (2, 2), (3, 2), (4, 2),
                                 (5, 5), (6, 6), (7, 6), (8, 6), (9, 6),
                                 (10, 7), (12.99, 7), (13, 9), (15, 10),
                                 (16, 10)]:
            self.set_edit(widget.threshold_edit, threshold)
            self.assertEqual(float(widget.edges_edit.text()), edges,
                             msg=f"at threshold={threshold}")

    def test_edges_from_threshold_asymmetric(self):
        widget = self.widget
        self.distances[0, 1] = 1.5
        self.send_signal(widget.Inputs.distances, self.distances)

        for threshold, edges in [(0, 0), (1, 1), (1.5, 2), (2, 4), (3, 4), (4, 4),
                                 (5, 10), (6, 12), (7, 12), (8, 12), (9, 12),
                                 (10, 14), (12.99, 14), (13, 18), (15, 20),
                                 (16, 20)]:
            self.set_edit(widget.threshold_edit, threshold)
            self.assertEqual(float(widget.edges_edit.text()), edges,
                             msg=f"at threshold={threshold}")

    def test_set_edges(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)

        for edges, threshold, density in [(0, 0, 0), (1, 1, 10), (2, 2, 20),
                                          (3, 5, 50), (4, 5, 50), (5, 5, 50),
                                          (6, 6, 60), (7, 10, 70),
                                          (8, 13, 90), (9, 13, 90), (10, 15, 100)]:
            self.set_edit(widget.edges_edit, edges)
            self.assertEqual(widget.threshold, threshold,
                             msg=f"at edges={edges}")
            self.assertEqual(widget.edges_edit.text(), str(density // 10),
                             msg=f"at edges={edges}")
            self.assertEqual(widget.density, density,
                             msg=f"at edges={edges}")
            self.assertEqual(widget.histogram.hline.value(), density // 10,
                             msg=f"at edges={edges}")
            self.assertEqual(widget.histogram.vline.value(),
                             min(threshold, np.max(self.distances)),
                             msg=f"at edges={edges}")

    def test_set_density(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)

        for density, threshold, edges in [(0, 0, 0), (10, 1, 1), (20, 2, 2),
                                          (30, 5, 5), (40, 5, 5), (50, 5, 5),
                                          (60, 6, 6), (70, 10, 7),
                                          (80, 13, 9), (90, 13, 9), (100, 15, 10)]:
            self.set_edit(widget.density_edit, density)
            self.assertEqual(widget.threshold, threshold,
                             msg=f"at density={density}")
            self.assertEqual(int(widget.edges_edit.text()), edges,
                             msg=f"at density={density}")
            self.assertEqual(widget.density, edges * 10)
            self.assertEqual(widget.histogram.hline.value(), edges,
                             msg=f"at threshold={threshold}")
            self.assertEqual(widget.histogram.vline.value(),
                             min(threshold, np.max(self.distances)),
                             msg=f"at threshold={threshold}")

    def test_set_threshold(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)

        for threshold, density, edges in [(0, 0, 0), (1, 10, 1), (2, 20, 2),
                                          (3, 20, 2), (4, 20, 2), (5, 50, 5),
                                          (6, 60, 6), (9, 60, 6), (10, 70, 7),
                                          (12.9, 70, 7), (13, 90, 9),
                                          (14.9, 90, 9), (15, 100, 10),
                                          (16, 100, 10)]:
            self.set_edit(widget.threshold_edit, threshold)
            self.assertEqual(widget.density, density,
                             msg=f"at threshold={threshold}")
            self.assertEqual(int(widget.edges_edit.text()), edges,
                             msg=f"at threshold={threshold}")
            self.assertEqual(widget.histogram.hline.value(), edges,
                             msg=f"at threshold={threshold}")
            self.assertEqual(widget.histogram.vline.value(),
                             min(threshold, np.max(self.distances)),
                             msg=f"at threshold={threshold}")

    @patch("orangecontrib.network.widgets.OWNxFromDistances.OWNxFromDistances.generate_network")
    def test_set_threshold_from_histogram(self, generate_network):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)

        generate_network.reset_mock()
        widget.histogram.thresholdChanged.emit(5)
        generate_network.assert_not_called()
        self.assertEqual(widget.threshold, 5)
        self.assertEqual(widget.density, 50)
        self.assertEqual(widget.edges_edit.text(), "~5")

        widget.histogram.thresholdChanged.emit(11)
        generate_network.assert_not_called()
        self.assertEqual(widget.threshold, 11)
        self.assertEqual(widget.density, 70)
        self.assertEqual(widget.edges_edit.text(), "~7")

        widget.histogram.thresholdChanged.emit(15)
        generate_network.assert_not_called()
        self.assertEqual(widget.threshold, 15)
        self.assertEqual(widget.density, 100)
        self.assertEqual(widget.edges_edit.text(), "~10")

        widget.histogram.thresholdChanged.emit(16)
        generate_network.assert_not_called()
        self.assertEqual(widget.threshold, 16)
        self.assertEqual(widget.density, 100)
        self.assertEqual(widget.edges_edit.text(), "~10")

        widget.histogram.draggingFinished.emit()
        generate_network.assert_called_once()

    def test_generate_network(self):
        widget = self.widget

        distances = DistMatrix(np.array([[0., 1, 2, 5, 10],
                                         [1, -1, 5, 5, 3],
                                         [2, 5, 2, 6, 13],
                                         [5, 5, 6, 3, 15],
                                         [10, 3, 13, 15, 0]]))

        widget.threshold = 2
        self.send_signal(widget.Inputs.distances, distances)
        graph = self.get_output(widget.Outputs.network)
        self.assertEqual(graph.number_of_nodes(), 5)
        self.assertEqual(graph.number_of_edges(), 2)
        coo = graph.edges[0].edges.tocoo()
        np.testing.assert_equal(coo.row, [1, 2])
        np.testing.assert_equal(coo.col, [0, 0])
        np.testing.assert_almost_equal(coo.data, [1 / 10 ** 0.5, 0.1])
        self.assertEqual(list(graph.nodes), list("12345"))

        self.set_edit(widget.threshold_edit, 3)
        graph = self.get_output(widget.Outputs.network)
        self.assertEqual(graph.number_of_nodes(), 5)
        self.assertEqual(graph.number_of_edges(), 3)
        coo = graph.edges[0].edges.tocoo()
        np.testing.assert_equal(coo.row, [1, 2, 4])
        np.testing.assert_equal(coo.col, [0, 0, 1])
        np.testing.assert_almost_equal(coo.data, [1 / 10 ** (1 / 3), 1 / 10 ** (2 / 3), 0.1])

        self.set_edit(widget.threshold_edit, 4)
        graph = self.get_output(widget.Outputs.network)
        self.assertEqual(graph.number_of_nodes(), 5)
        self.assertEqual(graph.number_of_edges(), 3)
        coo = graph.edges[0].edges.tocoo()
        np.testing.assert_equal(coo.row, [1, 2, 4])
        np.testing.assert_equal(coo.col, [0, 0, 1])
        np.testing.assert_almost_equal(coo.data, [1 / 10 ** (1 / 3), 1 / 10 ** (2 / 3), 0.1])

        self.set_edit(widget.threshold_edit, 5)
        graph = self.get_output(widget.Outputs.network)
        self.assertEqual(graph.number_of_nodes(), 5)
        self.assertEqual(graph.number_of_edges(), 6)
        coo = graph.edges[0].edges.tocoo()
        np.testing.assert_equal(coo.row, [1, 2, 2, 3, 3, 4])
        np.testing.assert_equal(coo.col, [0, 0, 1, 0, 1, 1])
        np.testing.assert_almost_equal(coo.data, [1 / 10 ** (1 / 5), 1 / 10 ** (2 / 5),
                                                  0.1, 0.1, 0.1, 1 / 10 ** (3 / 5)])

    def test_threshold_decimals(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)

        widget.threshold = 1.23412
        widget.update_edits(widget.edges_edit)
        self.assertEqual(widget.threshold_edit.text(), "1.2")

        self.send_signal(widget.Inputs.distances, self.distances / 100)
        widget.threshold = 1.23412
        widget.update_edits(widget.edges_edit)
        self.assertEqual(widget.threshold_edit.text(), "1.234")

    def test_too_many_edges(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)

        with patch("orangecontrib.network.widgets.OWNxFromDistances."
                   "OWNxFromDistances.edges_from_threshold",
                   return_value=1_000_000):
            widget.generate_network()
        self.assertTrue(widget.Error.number_of_edges.is_shown())

        widget.generate_network()
        self.assertFalse(widget.Error.number_of_edges.is_shown())

    def test_report(self):
        widget = self.widget
        widget.send_report()

        self.send_signal(widget.Inputs.distances, self.distances)
        widget.send_report()


class TestHistogram(GuiTest):
    def setUp(self):
        self.histogram = Histogram(None)
        assert set(self.histogram._elements()) == \
               {self.histogram.hline, self.histogram.vline,
                self.histogram.curve, self.histogram.fill_curve}

    def test_show_hide_elements(self):
        hist = self.histogram

        self.assertFalse(any(x.isVisible() for x in hist._elements()))
        hist.set_graph(np.array([1, 2, 5, 6, 10, 13, 15]),
                       np.array([1, 2, 5, 6, 7, 9, 10]))
        self.assertTrue(all(x.isVisible() for x in hist._elements()))

        hist.clear_graph()
        self.assertFalse(any(x.isVisible() for x in hist._elements()))

    def test_ranges(self):
        hist = self.histogram

        hist.set_graph(np.array([1, 2, 5, 6, 10, 13, 15]),
                       np.array([1, 2, 5, 6, 7, 9, 10]))
        self.assertEqual(hist.hline.bounds(), [0, 10])
        self.assertEqual(hist.vline.bounds(), [0, 15])
        self.assertEqual(hist.getAxis("left").range, [0, 10])
        self.assertEqual(hist.getAxis("bottom").range, [0, 15])
        self.assertAlmostEqual(hist.prop_axis.scale, 1 / 10 * 100)

    def _assert_fill_curve(self, hist, data):
        for exp, act in zip(data, hist.fill_curve.getData()):
            np.testing.assert_almost_equal(exp, act)

    def test_drag_hline(self):
        hist = self.histogram
        spy = QSignalSpy(hist.thresholdChanged)

        hist.set_graph(np.array([1, 2, 5, 6, 10, 13, 15]),
                       np.array([1, 2, 5, 6, 7, 9, 10]))

        hist.hline.setPos(1)
        hist.hline.sigDragged.emit(hist.hline)
        self.assertEqual(hist.vline.value(), 1)
        self.assertEqual(list(spy), [[1]])
        self._assert_fill_curve(hist, ([0, 1], [0, 1]))

        hist.hline.setPos(1.5)
        hist.hline.sigDragged.emit(hist.hline)
        self.assertEqual(hist.vline.value(), 2)
        self.assertEqual(list(spy), [[1], [2]])
        self._assert_fill_curve(hist, ([0, 1, 2], [0, 1, 2]))

        hist.hline.setPos(6.8)
        hist.hline.sigDragged.emit(hist.hline)
        self.assertEqual(hist.vline.value(), 10)
        self.assertEqual(list(spy), [[1], [2], [10]])
        self._assert_fill_curve(hist,
                                ([0, 1, 2, 5, 6, 10],
                                 [0, 1, 2, 5, 6, 7]))

        hist.hline.setPos(7)
        hist.hline.sigDragged.emit(hist.hline)
        self.assertEqual(hist.vline.value(), 10)
        self.assertEqual(list(spy), [[1], [2], [10], [10]])
        self._assert_fill_curve(hist,
                                ([0, 1, 2, 5, 6, 10],
                                 [0, 1, 2, 5, 6, 7]))

        hist.hline.setPos(7.2)
        hist.hline.sigDragged.emit(hist.hline)
        self.assertEqual(hist.vline.value(), 13)
        self.assertEqual(list(spy), [[1], [2], [10], [10], [13]])
        self._assert_fill_curve(hist,
                                ([0, 1, 2, 5, 6, 10, 13],
                                 [0, 1, 2, 5, 6, 7, 9]))

        hist.hline.setPos(10)
        hist.hline.sigDragged.emit(hist.hline)
        self.assertEqual(hist.vline.value(), 15)
        self.assertEqual(list(spy), [[1], [2], [10], [10], [13], [15]])
        self._assert_fill_curve(hist,
                                ([0, 1, 2, 5, 6, 10, 13, 15],
                                 [0, 1, 2, 5, 6, 7, 9, 10]))

        hist.hline.setPos(0)
        hist.hline.sigDragged.emit(hist.hline)
        self.assertEqual(hist.vline.value(), 0)
        self.assertEqual(list(spy), [[1], [2], [10], [10], [13], [15], [0]])
        self._assert_fill_curve(hist,
                                ([0],
                                 [0]))

    def test_drag_vline(self):
        hist = self.histogram
        spy = QSignalSpy(hist.thresholdChanged)

        hist.set_graph(np.array([1, 2, 5, 6, 10, 13, 15]),
                       np.array([1, 2, 5, 6, 7, 9, 10]))

        hist.vline.setPos(1)
        hist.vline.sigDragged.emit(hist.vline)
        self.assertEqual(hist.hline.value(), 1)
        self.assertEqual(list(spy), [[1]])
        self._assert_fill_curve(hist, ([0, 1], [0, 1]))

        hist.vline.setPos(1.5)
        hist.vline.sigDragged.emit(hist.vline)
        self.assertEqual(hist.hline.value(), 1)
        self.assertEqual(list(spy), [[1], [1.5]])
        self._assert_fill_curve(hist, ([0, 1], [0, 1]))

        hist.vline.setPos(2)
        hist.vline.sigDragged.emit(hist.vline)
        self.assertEqual(hist.hline.value(), 2)
        self.assertEqual(list(spy), [[1], [1.5], [2]])
        self._assert_fill_curve(hist, ([0, 1, 2], [0, 1, 2]))

        hist.vline.setPos(9)
        hist.vline.sigDragged.emit(hist.vline)
        self.assertEqual(hist.hline.value(), 6)
        self.assertEqual(list(spy), [[1], [1.5], [2], [9]])
        self._assert_fill_curve(hist,
                                ([0, 1, 2, 5, 6],
                                 [0, 1, 2, 5, 6]))

        hist.vline.setPos(10)
        hist.vline.sigDragged.emit(hist.vline)
        self.assertEqual(hist.hline.value(), 7)
        self.assertEqual(list(spy), [[1], [1.5], [2], [9], [10]])
        self._assert_fill_curve(hist,
                                ([0, 1, 2, 5, 6, 10],
                                 [0, 1, 2, 5, 6, 7]))

        hist.vline.setPos(10.5)
        hist.vline.sigDragged.emit(hist.vline)
        self.assertEqual(hist.hline.value(), 7)
        self.assertEqual(list(spy), [[1], [1.5], [2], [9], [10], [10.5]])
        self._assert_fill_curve(hist,
                                ([0, 1, 2, 5, 6, 10],
                                 [0, 1, 2, 5, 6, 7]))

        hist.vline.setPos(13)
        hist.vline.sigDragged.emit(hist.vline)
        self.assertEqual(hist.hline.value(), 9)
        self.assertEqual(list(spy), [[1], [1.5], [2], [9], [10], [10.5], [13]])
        self._assert_fill_curve(hist,
                                ([0, 1, 2, 5, 6, 10, 13],
                                 [0, 1, 2, 5, 6, 7, 9]))

        hist.vline.setPos(0)
        hist.vline.sigDragged.emit(hist.vline)
        self.assertEqual(hist.hline.value(), 0)
        self.assertEqual(list(spy), [[1], [1.5], [2], [9], [10], [10.5], [13], [0]])
        self._assert_fill_curve(hist,
                                ([0],
                                 [0]))


class TestQIntValidatorWithFixup(GuiTest):
    def test_validator(self):
        def enter_text(t):
            e.setText(t)
            e.keyPressEvent(QKeyEvent(QEvent.KeyPress, Qt.Key_Enter, Qt.NoModifier))

        e = QLineEdit()
        e.setValidator(QIntValidatorWithFixup(0, 100, e))
        enter_text("")
        self.assertEqual(e.text(), "")
        enter_text("1")
        self.assertEqual(e.text(), "1")
        enter_text("100")
        self.assertEqual(e.text(), "100")
        enter_text("101")
        self.assertEqual(e.text(), "100")


if __name__ == "__main__":
    unittest.main()



import numpy as np
import scipy.sparse as sp

from AnyQt.QtCore import QTimer, QSize, Qt

import Orange
from Orange.data import Table, Domain, StringVariable
from Orange.util import allot
from Orange.widgets import gui
from Orange.widgets.settings import Setting, SettingProvider
from Orange.widgets.utils.concurrent import ConcurrentWidgetMixin
from Orange.widgets.utils.plot import OWPlotGUI
from Orange.widgets.visualize.utils.widget import OWDataProjectionWidget
from Orange.widgets.widget import Input, Output

from orangecontrib.network.network.base import Network
from orangecontrib.network.network.layout import fruchterman_reingold
from orangecontrib.network.widgets.graphview import GraphView
from orangewidget.widget import Message, Msg

FR_ALLOWED_TIME = 30


def run(positions, edges, observe_weights, init_temp, k, state):
    def update(positions, progress):
        state.set_progress_value(progress)
        if not large_graph:
            state.set_partial_result(positions)
        if state.is_interruption_requested():
            raise Exception  # pylint: disable=broad-exception-raised

    nnodes = positions.shape[0]
    large_graph = nnodes + edges.shape[0] > 30000
    sample_ratio = None if nnodes < 1000 else 1000 / nnodes
    fruchterman_reingold(
        positions, edges, observe_weights,
        FR_ALLOWED_TIME, k, init_temp, sample_ratio,
        callback=update)
    return positions


class OWNxExplorer(OWDataProjectionWidget, ConcurrentWidgetMixin):
    name = "Network Explorer"
    description = "Visually explore the network and its properties."
    icon = "icons/NetworkExplorer.svg"
    priority = 6420

    class Inputs:
        node_data = Input("Node Data", Table)
        node_subset = Input("Node Subset", Table)
        network = Input("Network", Network, default=True)
        node_distances = Input("Node Distances", Orange.misc.DistMatrix)

    class Outputs(OWDataProjectionWidget.Outputs):
        subgraph = Output("Selected sub-network", Network)
        unselected_subgraph = Output("Remaining sub-network", Network)
        distances = Output("Distance matrix", Orange.misc.DistMatrix)

    UserAdviceMessages = [
        Message("Double clicks select connected components",
                Message.Information),
    ]

    GRAPH_CLASS = GraphView
    graph = SettingProvider(GraphView)

    layout_density = Setting(10)
    observe_weights = Setting(True)

    mark_hops = Setting(1)
    mark_min_conn = Setting(5)
    mark_max_conn = Setting(5)
    mark_most_conn = Setting(1)

    alpha_value = 255  # Override the setting from parent

    class Warning(OWDataProjectionWidget.Warning):
        distance_matrix_mismatch = Msg(
            "Distance matrix size doesn't match the number of network nodes "
            "and will be ignored.")
        no_graph_found = Msg("Node data is given, graph data is missing.")

    class Error(OWDataProjectionWidget.Error):
        data_size_mismatch = Msg(
            "Length of the data does not match the number of nodes.")
        network_too_large = Msg("Network is too large to visualize.")

    def __init__(self):
        # These are already needed in super().__init__()
        self.number_of_nodes = 0
        self.number_of_edges = 0
        self.nHighlighted = 0
        self.nSelected = 0
        self.nodes_per_edge = 0
        self.edges_per_node = 0

        self.mark_mode = 0
        self.mark_text = ""

        OWDataProjectionWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        self.network = None
        self.node_data = None
        self.distance_matrix = None
        self.edges = None
        self.positions = None

        self.marked_nodes = None
        self.searchStringTimer = QTimer(self)
        self.searchStringTimer.timeout.connect(self.update_marks)
        self.set_mark_mode()
        self.setMinimumWidth(600)

    def sizeHint(self):
        return QSize(800, 600)

    def _add_controls(self):
        # pylint: disable=attribute-defined-outside-init
        self.gui = OWPlotGUI(self)
        self._add_info_box()
        self.gui.point_properties_box(self.controlArea)
        self._add_effects_box()
        self.gui.plot_properties_box(self.controlArea)
        self._add_mark_box()
        self.controls.attr_label.activated.connect(self.on_change_label_attr)

    def _add_info_box(self):
        # pylint: disable=attribute-defined-outside-init
        info = gui.vBox(self.controlArea, box="Layout")
        gui.label(
            info, self,
            "Nodes: %(number_of_nodes)i (%(nodes_per_edge).2f per edge); "
            "%(nSelected)i selected")
        gui.label(
            info, self,
            "Edges: %(number_of_edges)i (%(edges_per_node).2f per node)")
        lbox = gui.hBox(info)
        self.relayout_button = gui.button(
            lbox, self, 'Improve', callback=self.improve, autoDefault=False,
            tooltip="Optimize the current layout, with a small initial jerk")
        self.stop_button = gui.button(
            lbox, self, 'Stop', callback=self.stop_relayout, autoDefault=False,
            hidden=True)
        self.randomize_button = gui.button(
            lbox, self, 'Re-layout', callback=self.restart, autoDefault=False,
            tooltip="Restart laying out from random positions")
        gui.hSlider(info, self, "layout_density", minValue=1, maxValue=50,
                    label="Gravity", orientation=Qt.Horizontal,
                    callback_finished=self.improve,
                    tooltip="Lower values improve optimization,\n"
                            "higher work better for graph with many small "
                            "components")
        gui.checkBox(info, self, "observe_weights",
                     label="Make edges with large weights shorter",
                     callback=self.improve)

    def _add_effects_box(self):
        # pylint: disable=attribute-defined-outside-init
        gbox = self.gui.create_gridbox(self.controlArea, box="Widths and Sizes")
        self.gui.add_widget(self.gui.PointSize, gbox)
        gbox.layout().itemAtPosition(1, 0).widget().setText("Node Size:")
        self.gui.add_control(
            gbox, gui.hSlider, "Edge width:",
            master=self, value='graph.edge_width',
            minValue=1, maxValue=10, step=1,
            callback=self.graph.update_edges)
        box = gui.vBox(None)
        gbox.layout().addWidget(box, 3, 0, 1, 2)
        gui.separator(box)
        self.checkbox_relative_edges = gui.checkBox(
            box, self, 'graph.relative_edge_widths',
            'Scale edge widths to weights',
            callback=self.graph.update_edges)
        self.checkbox_show_weights = gui.checkBox(
            box, self, 'graph.show_edge_weights',
            'Show edge weights',
            callback=self.graph.update_edge_labels)
        self.checkbox_show_weights = gui.checkBox(
            box, self, 'graph.label_selected_edges',
            'Label only edges of selected nodes',
            callback=self.graph.update_edge_labels)

        # This is ugly: create a slider that controls alpha_value so that
        # parent can enable and disable it - although it's never added to any
        # layout and visible to the user
        gui.hSlider(None, self, "graph.alpha_value")

    def _add_mark_box(self):
        # pylint: disable=attribute-defined-outside-init
        hbox = gui.hBox(None, box=True)
        self.mainArea.layout().addWidget(hbox)
        vbox = gui.hBox(hbox)

        def spin(value, label, minv, maxv):
            return gui.spin(
                vbox, self, value, label=label, minv=minv, maxv=maxv,
                step=1,
                alignment=Qt.AlignRight, callback=self.update_marks).box

        def text_line():
            def set_search_string_timer():
                self.searchStringTimer.stop()
                self.searchStringTimer.start(300)

            return gui.lineEdit(
                gui.hBox(vbox), self, "mark_text", label="Text: ",
                orientation=Qt.Horizontal, minimumWidth=50,
                callback=set_search_string_timer, callbackOnType=True).box

        def _mark_by_labels(marker):
            txt = self.mark_text.lower()
            if not txt:
                return None
            labels = self.get_label_data()
            if labels is None:
                return None
            return marker(np.char.array(labels), txt)

        def mark_label_equals():
            return _mark_by_labels(
                lambda labels, txt: np.flatnonzero(labels.lower() == txt))

        def mark_label_starts():
            return _mark_by_labels(
                lambda labels, txt: np.flatnonzero(labels.lower().startswith(txt)))

        def mark_label_contains():
            return _mark_by_labels(
                lambda labels, txt: np.flatnonzero(labels.lower().find(txt) != -1))

        def mark_text():
            txt = self.mark_text.lower()
            if not txt or self.data is None:
                return None
            return np.array(
                [i for i, inst in enumerate(self.data)
                 if txt in "\x00".join(map(str, inst.list)).lower()])

        def mark_reachable():
            selected = self.graph.get_selection()
            if selected is None:
                return None
            return self.get_reachable(selected)

        def mark_close():
            selected = self.graph.get_selection()
            if selected is None:
                return None
            neighbours = set(selected)
            last_round = list(neighbours)
            for _ in range(self.mark_hops):
                next_round = set()
                for neigh in last_round:
                    next_round |= set(self.network.neighbours(neigh))
                neighbours |= next_round
                last_round = next_round
            neighbours -= set(selected)
            return np.array(list(neighbours))

        def mark_from_input():
            if self.subset_data is None or self.data is None:
                return None
            ids = set(self.subset_data.ids)
            return np.array(
                [i for i, ex in enumerate(self.data) if ex.id in ids])

        def mark_most_connections():
            n = self.mark_most_conn
            if n >= self.number_of_nodes:
                return np.arange(self.number_of_nodes)
            degrees = self.network.degrees()
            # pylint: disable=invalid-unary-operand-type
            min_degree = np.partition(degrees, -n)[-n]
            return np.flatnonzero(degrees >= min_degree)

        def mark_more_than_any_neighbour():
            degrees = self.network.degrees()
            return np.array(
                [node for node, degree in enumerate(degrees)
                 if degree > np.max(degrees[self.network.neighbours(node)],
                                    initial=0)])

        def mark_more_than_average_neighbour():
            degrees = self.network.degrees()
            return np.array(
                [node for node, degree, neighbours in (
                    (node, degree, self.network.neighbours(node))
                     for node, degree in enumerate(degrees))
                 if degree > (np.mean(degrees[neighbours]) if neighbours.size else 0)
                 ]
            )

        self.mark_criteria = [
            ("(Select criteria for marking)", None, lambda: np.zeros((0,))),
            ("Mark node labelled as", text_line(), mark_label_equals),
            ("Mark nodes whose label starts with", text_line(), mark_label_starts),
            ("Mark nodes whose label contains", text_line(), mark_label_contains),
            ("Mark nodes whose data contains", text_line(), mark_text),
            ("Mark nodes reachable from selected", None, mark_reachable),

            ("Mark nodes in vicinity of selection",
             spin("mark_hops", "Number of hops:", 1, 20),
             mark_close),

            ("Mark nodes from subset signal", None, mark_from_input),

            ("Mark nodes with few connections",
             spin("mark_max_conn", "Max. connections:", 0, 1000),
             lambda: np.flatnonzero(self.network.degrees() <= self.mark_max_conn)),

            ("Mark nodes with many connections",
             spin("mark_min_conn", "Min. connections:", 1, 1000),
             lambda: np.flatnonzero(self.network.degrees() >= self.mark_min_conn)),

            ("Mark nodes with most connections",
             spin("mark_most_conn", "Number of marked:", 1, 1000),
             mark_most_connections),

            ("Mark nodes with more connections than any neighbour", None,
             mark_more_than_any_neighbour),

            ("Mark nodes with more connections than average neighbour", None,
             mark_more_than_average_neighbour)
        ]

        cb = gui.comboBox(
            hbox, self, "mark_mode",
            items=[item for item, *_ in self.mark_criteria],
            maximumContentsLength=-1, callback=self.set_mark_mode)
        hbox.layout().insertWidget(0, cb)

        gui.rubber(hbox)
        self.btselect = gui.button(
            hbox, self, "Select", callback=self.select_marked)
        self.btadd = gui.button(
            hbox, self, "Add to Selection", callback=self.select_add_marked)
        self.btgroup = gui.button(
            hbox, self, "Add New Group", callback=self.select_as_group)

    def set_mark_mode(self, mode=None):
        if mode is not None:
            self.mark_mode = mode
        for i, (_, widget, _) in enumerate(self.mark_criteria):
            if widget:
                if i == self.mark_mode:
                    widget.show()
                else:
                    widget.hide()
        self.searchStringTimer.stop()
        self.update_marks()

    def update_marks(self):
        if self.network is None:
            return
        self.marked_nodes = self.mark_criteria[self.mark_mode][2]()
        if self.marked_nodes is not None and not self.marked_nodes.size:
            self.marked_nodes = None
        self.graph.update_marks()
        if self.graph.label_only_selected:
            self.graph.update_labels()
        self.update_selection_buttons()

    def update_selection_buttons(self):
        if self.marked_nodes is None:
            self.btselect.hide()
            self.btadd.hide()
            self.btgroup.hide()
            return
        else:
            self.btselect.show()

        selection = self.graph.get_selection()
        if len(selection) == 0 or np.max(selection) == 0:
            self.btadd.hide()
            self.btgroup.hide()
        elif np.max(selection) == 1:
            self.btadd.setText("Add to Selection")
            self.btadd.show()
            self.btgroup.hide()
        else:
            self.btadd.setText("Add to Group")
            self.btadd.show()
            self.btgroup.show()

    def selection_changed(self):
        super().selection_changed()
        self.nSelected = 0 if self.selection is None else len(self.selection)
        self.update_selection_buttons()
        self.update_marks()

    def select_marked(self):
        self.graph.selection_select(self.marked_nodes)

    def select_add_marked(self):
        self.graph.selection_append(self.marked_nodes)

    def select_as_group(self):
        self.graph.selection_new_group(self.marked_nodes)

    def on_change_label_attr(self):
        if self.mark_mode in (1, 2):
            self.update_marks()

    @Inputs.node_data
    def set_node_data(self, data):
        self.node_data = data

    @Inputs.node_subset
    def set_node_subset(self, data):
        # It would be better to call super, but this fails because super
        # is decorated to set the partial summary for signal "Subset Data",
        # which does not exist for this widget (OWNxExplorer.Inputs is not
        # derived from OWDataProjectionWidget.Inputs in order to rename the
        # signal)
        self.subset_data = data

    @Inputs.node_distances
    def set_items_distance_matrix(self, matrix):
        self.distance_matrix = matrix
        self.positions = None

    @Inputs.network
    def set_graph(self, graph):

        def set_graph_none(error=None):
            if error is not None:
                error()
            self.network = None
            self.number_of_nodes = self.edges_per_node = 0
            self.number_of_edges = self.nodes_per_edge = 0

        def compute_stats():
            self.number_of_nodes = graph.number_of_nodes()
            self.number_of_edges = graph.number_of_edges()
            self.edges_per_node = self.number_of_edges / self.number_of_nodes
            self.nodes_per_edge = \
                self.number_of_nodes / max(1, self.number_of_edges)

        self.mark_text = ""
        self.set_mark_mode(0)
        self.positions = None

        if not graph or graph.number_of_nodes() == 0:
            set_graph_none()
            return
        if graph.number_of_nodes() + graph.number_of_edges() > 100000:
            set_graph_none(self.Error.network_too_large)
            return
        self.Error.clear()

        self.network = graph
        compute_stats()

    def handleNewSignals(self):
        network = self.network

        def set_actual_data():
            self.closeContext()
            self.Error.data_size_mismatch.clear()
            self.Warning.no_graph_found.clear()
            if network is None:
                if self.node_data is not None:
                    self.Warning.no_graph_found()
                return
            n_nodes = len(self.network.nodes)
            if self.node_data is not None:
                if len(self.node_data) != n_nodes:
                    self.Error.data_size_mismatch()
                    self.data = None
                else:
                    self.data = self.node_data
            if self.node_data is None:
                if isinstance(network.nodes, Table):
                    self.data = network.nodes
                elif isinstance(network.nodes, np.ndarray) \
                        and (len(network.nodes.shape) == 1
                             or network.nodes.shape[1] == 1):
                    self.data = Table.from_numpy(
                        Domain([], None, [StringVariable("label")]),
                        np.zeros((len(network.nodes),0)),
                        None,
                        metas=network.nodes.reshape((n_nodes, 1))
                    )
                else:
                    self.data = None

            if self.data is not None:
                # Replicate the necessary parts of set_data
                self.valid_data = np.full(len(self.data), True, dtype=bool)
                self.init_attr_values()
                self.openContext(self.data)
                self.cb_class_density.setEnabled(self.can_draw_density())

        def set_actual_edges():
            def set_checkboxes(value):
                self.checkbox_show_weights.setEnabled(value)
                self.checkbox_relative_edges.setEnabled(value)

            self.Warning.distance_matrix_mismatch.clear()

            if self.network is None:
                self.edges = None
                set_checkboxes(False)
                return

            set_checkboxes(True)
            if network.number_of_edges(0):
                self.edges = network.edges[0].edges.tocoo()
            else:
                self.edges = sp.coo_matrix((0, 3))
            if self.distance_matrix is not None:
                if len(self.distance_matrix) != self.number_of_nodes:
                    self.Warning.distance_matrix_mismatch()
                else:
                    self.edges.data = np.fromiter(
                        (self.distance_matrix[u, v]
                         for u, v in zip(self.edges.row, self.edges.col)),
                        dtype=np.int32, count=len(self.edges.row)
                    )
            if np.allclose(self.edges.data, 0):
                self.edges.data[:] = 1
                set_checkboxes(False)
            elif len(set(self.edges.data)) == 1:
                set_checkboxes(False)

        self.cancel()
        set_actual_data()
        super()._handle_subset_data()
        if self.positions is None:
            set_actual_edges()
            self.set_random_positions()
            self.graph.reset_graph()
            self.relayout(True)
        else:
            self.graph.update_point_props()
        self.update_marks()
        self.update_selection_buttons()

    def init_attr_values(self):
        super().init_attr_values()
        if self.node_data is None \
                and self.data is not None \
                and isinstance(self.network.nodes, np.ndarray):
            assert len(self.data.domain.metas) == 1
            self.attr_label = self.data.domain.metas[0]

    def randomize(self):
        self.set_random_positions()
        self.graph.update_coordinates()

    def set_random_positions(self):
        if self.network is None:
            self.positions = None
        else:
            self.positions = np.random.uniform(size=(self.number_of_nodes, 2))

    def get_reachable(self, initial):
        to_check = list(initial)
        reachable = set(to_check)
        for node in to_check:
            new_checks = set(self.network.neighbours(node)) - reachable
            to_check += new_checks
            reachable |= new_checks
        return np.array(to_check)

    def send_data(self):
        super().send_data()

        Outputs = self.Outputs
        selected_indices = self.graph.get_selection()
        if selected_indices is None or len(selected_indices) == 0:
            Outputs.subgraph.send(None)
            Outputs.unselected_subgraph.send(self.network)
            Outputs.distances.send(None)
            return

        selection = self.graph.selection
        subgraph = self.network.subgraph(selected_indices)
        subgraph.nodes = \
            self._get_selected_data(self.data, selected_indices, selection)
        Outputs.subgraph.send(subgraph)
        Outputs.unselected_subgraph.send(
            self.network.subgraph(np.flatnonzero(selection == 0)))
        distances = self.distance_matrix
        if distances is None:
            Outputs.distances.send(None)
        else:
            Outputs.distances.send(distances.submatrix(sorted(selected_indices)))

    def get_coordinates_data(self):
        if self.positions is not None:
            return self.positions.T
        else:
            return None, None

    def get_embedding(self):
        return self.positions

    def get_subset_mask(self):
        if self.data is None:
            return None
        return super().get_subset_mask()

    def get_edges(self):
        return self.edges

    def is_directed(self):
        return self.network is not None and self.network.edges[0].directed

    def get_marked_nodes(self):
        return self.marked_nodes

    def set_buttons(self, running):
        self.stop_button.setHidden(not running)
        self.relayout_button.setHidden(running)
        self.randomize_button.setHidden(running)

    def stop_relayout(self):
        self.cancel()

    def restart(self):
        self.relayout(restart=True)

    def improve(self):
        self.relayout(restart=False)

    def relayout(self, restart):
        if self.edges is None:
            return
        if restart or self.positions is None:
            self.set_random_positions()

        Simplifications = self.graph.Simplifications
        self.graph.set_simplifications(
            Simplifications.NoDensity
            + Simplifications.NoLabels * (len(self.graph.labels) > 20)
            + Simplifications.NoEdgeLabels * (len(self.graph.edge_labels) > 20)
            + Simplifications.NoEdges * (self.number_of_edges > 1000))

        init_temp = 0.05 if restart else 0.2
        k = self.layout_density / 10 / np.sqrt(self.number_of_nodes)
        self.set_buttons(running=True)
        self.start(run, self.positions, self.edges, self.observe_weights, init_temp, k)

    def cancel(self):
        self.set_buttons(running=False)
        super().cancel()

    def on_done(self, positions):  # pylint: disable=arguments-renamed
        self.positions = positions
        self.set_buttons(running=False)
        self.graph.set_simplifications(
            self.graph.Simplifications.NoSimplifications)
        self.graph.update_coordinates()

    @allot(0.02)
    def on_partial_result(self, positions):  # pylint: disable=arguments-renamed
        self.positions = positions
        self.graph.update_coordinates()

    def onDeleteWidget(self):
        self.shutdown()
        super().onDeleteWidget()

    def send_report(self):
        if self.network is None:
            return

        self.report_items('Graph info', [
            ("Number of vertices", self.network.number_of_nodes()),
            ("Number of edges", self.network.number_of_edges()),
            ("Vertices per edge", round(self.nodes_per_edge, 3)),
            ("Edges per vertex", round(self.edges_per_node, 3)),
        ])
        if self.data is not None:
            self.report_data("Data", self.data)
        if any((self.attr_color, self.attr_shape,
                self.attr_size, self.attr_label)):
            self.report_items(
                "Visual settings",
                [("Color", self._get_caption_var_name(self.attr_color)),
                 ("Label", self._get_caption_var_name(self.attr_label)),
                 ("Shape", self._get_caption_var_name(self.attr_shape)),
                 ("Size", self._get_caption_var_name(self.attr_size))])
        self.report_plot()


def main():
    # pylint: disable=import-outside-toplevel, unused-import
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    from orangecontrib.network.network.readwrite \
        import read_pajek, transform_data_to_orange_table
    from os.path import join, dirname

    network = read_pajek(join(dirname(dirname(__file__)), 'networks', 'leu_by_genesets.net'))
    # network = read_pajek(
    #    join(dirname(dirname(__file__)), 'networks', 'dicty_publication.net'))
    #network = read_pajek(join(dirname(dirname(__file__)), 'networks', 'davis.net'))
    #transform_data_to_orange_table(network)
    WidgetPreview(OWNxExplorer).run(set_graph=network)


if __name__ == "__main__":
    main()



import time

import numpy as np

from ._fr_layout import fruchterman_reingold_step

TEST_ITERATIONS = 10


def fruchterman_reingold(pos,
                         edges, weighted: bool,
                         allowed_time,
                         k,
                         init_temp=0.05,
                         sample_ratio=None,
                         callback_interval=0.5, callback=None):

    def run_iterations(n_iterations, temperatures, callback=callback):
        last_call = time.perf_counter()
        for iteration in range(n_iterations):
            if sample_ratio is not None:
                np.random.shuffle(sample)
            fruchterman_reingold_step(pos, sample[:sample_size],
                                      edge_src, edge_dst, edge_weights,
                                      k, temperatures[iteration], disp)

            if callback and \
                    (now := time.perf_counter()) - last_call > callback_interval:
                progress = iteration / n_iterations * 100
                callback(pos, progress)
                last_call = now

    n_nodes = len(pos)

    edge_weights = edges.data if weighted else np.empty(0)
    edge_src, edge_dst = edges.row.astype(np.int32), edges.col.astype(np.int32)

    disp = np.empty((n_nodes, 2))

    if sample_ratio is not None:
        sample_size = max(int(round(n_nodes * sample_ratio)), 1)
        sample = np.arange(n_nodes, dtype=np.int32)
    else:
        sample = np.zeros(0, dtype=np.int32)
        sample_size = 0

    temperatures = np.linspace(init_temp, 0.01, TEST_ITERATIONS)
    start_time = time.perf_counter()
    run_iterations(TEST_ITERATIONS, temperatures, callback=None)
    elapsed_time = time.perf_counter() - start_time

    iterations = int(allowed_time / (elapsed_time / TEST_ITERATIONS))
    if TEST_ITERATIONS > iterations:
        return
    iterations = min(iterations, 100 * n_nodes)
    temperatures = np.linspace(0.1, 0.015, iterations) ** 2
    run_iterations(iterations, temperatures)



import numpy as np
from Orange.data import Table, StringVariable, Domain


def items_from_distmatrix(matrix):
    if matrix.row_items is not None:
        row_items = matrix.row_items
        if isinstance(row_items, Table):
            if matrix.axis == 1:
                items = row_items
            else:
                items = [[v.name] for v in row_items.domain.attributes]
        else:
            items = [[str(x)] for x in matrix.row_items]
    else:
        items = [[str(i)] for i in range(1, 1 + matrix.shape[0])]
    if not isinstance(items, Table):
        items = Table.from_list(
            Domain([], metas=[StringVariable('label')]),
            items)
    return items


def weights_from_distances(weights):
    weights = weights.astype(np.float64)

    if weights.size == 0:
        return weights

    mi, ma = np.nanmin(weights), np.nanmax(weights)
    assert mi >= 0, "All distances must be positive"

    if ma - mi < 1e-6:
        return np.ones(weights.shape)

    weights /= ma
    weights *= np.log(10)
    np.exp(weights, out=weights)
    np.reciprocal(weights, out=weights)

    return weights



import unittest
from unittest.mock import Mock
from math import sqrt

import numpy as np
from scipy import sparse as sp

from Orange.data import Table, Domain, DiscreteVariable
from Orange.widgets.tests.base import simulate

from orangecontrib.network import Network
from orangecontrib.network.widgets.OWNxGroups import OWNxGroups
from orangecontrib.network.widgets.tests.utils import NetworkTest


class TestOWNxGroups(NetworkTest):
    def setUp(self):
        self.widget = self.create_widget(OWNxGroups)

    def test_inputs(self):
        Outputs = self.widget.Outputs
        network = self._read_network()
        table = self._read_items()

        # send network with items
        self.send_signal(self.widget.Inputs.network, network)
        names = [a.name for a in self.widget.effective_data.domain.attributes]
        self.assertListEqual(["department"], names)
        self.assertIsInstance(self.get_output(Outputs.network), Network)

        # send subset of items as data
        self.send_signal(self.widget.Inputs.data, table[:-1])
        self.assertTrue(self.widget.Error.data_size_mismatch.is_shown())
        self.assertIsNone(self.widget.effective_data)
        self.assertIsNone(self.get_output(self.widget.Outputs.network))

        # remove sent items
        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(self.widget.Error.data_size_mismatch.is_shown())
        names = [a.name for a in self.widget.effective_data.domain.attributes]
        self.assertListEqual(["department"], names)
        self.assertIsInstance(self.get_output(Outputs.network), Network)

        # send all items (with slightly different domain) as data
        self.send_signal(self.widget.Inputs.data, table)
        names = [a.name for a in self.widget.effective_data.domain.attributes]
        self.assertListEqual(["department"], names)
        self.assertIsInstance(self.get_output(Outputs.network), Network)

        # remove network
        self.send_signal(self.widget.Inputs.network, None)
        self.assertTrue(self.widget.Warning.no_graph_found.is_shown())
        self.assertIsNone(self.widget.effective_data)
        self.assertIsNone(self.get_output(self.widget.Outputs.network))

        # remove items
        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(self.widget.Warning.no_graph_found.is_shown())

    def test_outputs(self):
        in_net = self._read_network()
        self.send_signal(self.widget.Inputs.network, in_net)
        network = self.get_output(self.widget.Outputs.network)

        self.assertIsInstance(network, Network)
        self.assertIsInstance(network.nodes, Table)
        np.testing.assert_equal(set(network.nodes.X.flatten()), set(range(3)))
        self.assertEqual(
            set(zip(*network.edges[0].edges.nonzero())),
            {(0, 1), (1, 2), (0, 2)})

    def test_no_discrete_features(self):
        Outputs = self.widget.Outputs
        network = self._read_network("airtraffic.net")
        self.send_signal(self.widget.Inputs.network, network)
        self.assertIsInstance(self.get_output(Outputs.network), Network)

        network = self._read_network("mips_c2_cp_leu.net")
        self.send_signal(self.widget.Inputs.network, network)
        self.assertTrue(self.widget.Warning.no_discrete_features.is_shown())
        self.assertIsNone(self.get_output(Outputs.network))

        self.send_signal(self.widget.Inputs.network, None)
        self.assertFalse(self.widget.Warning.no_discrete_features.is_shown())

    def test_missing_values(self):
        network = self._read_network("airtraffic.net")
        self.send_signal(self.widget.Inputs.network, network)
        fname = network.nodes.domain.metas[-1].name
        simulate.combobox_activate_item(self.widget.controls.feature, fname)
        self.assertIsInstance(
            self.get_output(self.widget.Outputs.network), Network)

    def test_send_report(self):
        self.send_signal(self.widget.Inputs.network, self._read_network())
        self.widget.send_report()
        self.send_signal(self.widget.Inputs.network, None)
        self.widget.send_report()

    def test_disable_normalize(self):
        widget = self.widget
        buttons = widget.controls.weighting.buttons
        normalize = widget.controls.normalize

        buttons[widget.NoWeights].click()
        self.assertFalse(normalize.isEnabled())
        buttons[widget.WeightByDegrees].click()
        self.assertFalse(normalize.isEnabled())
        buttons[widget.WeightByWeights].click()
        self.assertTrue(normalize.isEnabled())
        buttons[widget.WeightByDegrees].click()
        self.assertFalse(normalize.isEnabled())

    def test_commit_on_normalization_change(self):
        widget = self.widget
        check = widget.controls.normalize
        widget.commit = Mock()

        check.click()
        widget.commit.assert_called()
        widget.commit.reset_mock()

        check.click()
        widget.commit.assert_called()

    def test_weights(self):
        label_var = DiscreteVariable("label", values=tuple("abcde"))
        domain = Domain([label_var], [])
        items = Table.from_numpy(domain, np.array([[0, 0, 1, 3, 4, 0, 1, 2, 4]]).T)
        src, dst, weights = np.array(
            [[0, 1, 5],
             [1, 2, 4],
             [3, 4, 6],
             [0, 5, 3],
             [1, 6, 1],
             [2, 7, 2],
             [3, 8, 8],
             [4, 8, 7],
             [5, 6, 2]]).T
        edges = sp.coo_matrix((weights.astype(float), (src, dst)), shape=(9, 9))
        network = Network(items, edges)
        expected = np.zeros((5, 5), dtype=float)

        widget = self.widget
        buttons = widget.controls.weighting.buttons
        self.send_signal(widget.Inputs.network, network)

        buttons[widget.NoWeights].click()
        groups = widget._map_network()
        for i, j in [(0, 1), (1, 2), (3, 4)]:
            expected[i, j] = 1
        np.testing.assert_equal(groups.edges[0].edges.todense(), expected)

        buttons[widget.WeightByDegrees].click()
        groups = widget._map_network()
        expected[0, 1] = 3
        expected[1, 2] = 1
        expected[3, 4] = 2
        np.testing.assert_equal(groups.edges[0].edges.todense(), expected)

        buttons[widget.WeightByWeights].click()
        widget.normalize = False
        groups = widget._map_network()
        expected[0, 1] = 7
        expected[1, 2] = 2
        expected[3, 4] = 14
        np.testing.assert_equal(groups.edges[0].edges.todense(), expected)

        widget.normalize = True
        groups = widget._map_network()
        expected[0, 1] = 1 / sqrt(10 * 3) + 4 / sqrt(10 * 6) + 2 / sqrt(5 * 3)
        expected[1, 2] = 2 / sqrt(6 * 2)
        expected[3, 4] = 6 / sqrt(14 * 13) + 8 / sqrt(14 * 15)
        np.testing.assert_equal(groups.edges[0].edges.todense(), expected)


if __name__ == "__main__":
    unittest.main()



from functools import wraps

import numpy as np
import scipy.sparse as sp
from sklearn.utils.extmath import row_norms

from orangecontrib.network import Network

__all__ = ["path", "cycle", "complete", "complete_bipartite", "barbell",
           "ladder", "circular_ladder", "grid", "hypercube",
           "lollipop", "star", "geometric"]

def from_csr(f):
    @wraps(f)
    def wrapped(*args):
        edges = f(*args)
        return Network(
            range(edges.shape[0]), edges,
            name=f"{f.__name__}{args}".replace(",)", ")"))
    return wrapped


def from_row_col(f):
    @wraps(f)
    def wrapped(*args):
        row, col, *n = f(*args)
        n = n[0] if n else max(np.max(row), np.max(col)) + 1
        edges = sp.csr_matrix((np.ones(len(row)), (row, col)), shape=(n, n))
        return Network(
            range(n), edges,
            name=f"{f.__name__}{args}".replace(",)", ")"))
    return wrapped


def from_dense(f):
    @wraps(f)
    def wrapped(*args):
        m = f(*args)
        return Network(
            range(len(m)), sp.csr_matrix(m),
            name=f"{f.__name__}{args}".replace(",)", ")"))
    return wrapped


@from_row_col
def path(n):
    return np.arange(n - 1), np.arange(n - 1) + 1, n


@from_row_col
def cycle(n):
    return np.arange(n), np.arange(1, n + 1) % n


@from_csr
def complete(n):
    if n < 5_000:
        return sp.triu(np.ones((n, n)), k=1, format="csr")

    # manually construct sparse matrix elements to avoid storing a dense (n x n) matrix in memory
    indptr = [0]
    for n_above_diag in range(n - 1, 0, -1):
        indptr.append(indptr[-1] + n_above_diag)
    indptr.append(indptr[-1])
    indptr = np.array(indptr)
    indices = np.concatenate([np.arange(row, n) for row in range(1, n)])
    data = np.ones(n * (n - 1) // 2)

    return sp.csr_matrix((data, indices, indptr), shape=(n, n))


@from_row_col
def complete_bipartite(n, m):
    return np.repeat(np.arange(n), m), n + np.arange(n * m) % m


@from_dense
def barbell(n, m):
    e = np.zeros((n + m, n + m))
    e[:n, :n] = 1  # complete
    e[-m:, -m:] = 1  # complete
    e[n, -m - 1] = e[n - 1, -m] = 1  # bridge
    e -= np.eye(n + m, n + m)  # no loops
    return np.triu(e)


def ladder(n):
    return grid(n, 2)


def circular_ladder(n):
    return grid(n, 2, True)


@from_csr
def grid(n, m=4, circular=False):
    t = n * m
    row = np.arange(2 * t - n) % t
    col = np.zeros(2 * t - n, dtype=int)
    data = np.ones(2 * t - n)

    col[:t] = np.arange(1, t + 1)  # horizontal edges
    col[n - 1:t:n] = np.arange(m)  # circle to prevent index out of range
    if not circular:
        data[n - 1:t:n] = 0  # eliminate circle

    col[t:] = np.arange(n, t)  # vertical edges
    edges = sp.csr_matrix((data, (row, col)), shape=(t, t))
    edges.eliminate_zeros()
    return edges


# https://math.stackexchange.com/questions/2328139/adjacency-matrix-for-n-dimensional-hypercube-graph
def _hypercube(ndim):
    """Recursively construct the edge-connectivity of a hypercube

    Parameters
    ----------
    ndim : int
        Dimension of the hypercube

    Returns
    -------
    ndarray, [2**ndim, 2**ndim], bool
        connectivity pattern of the hypercube
    """
    if ndim == 0:
        return np.array([[0]])
    else:
        d = _hypercube(ndim-1)
        i = np.eye(len(d))
        return np.block([
            [d, i],
            [i, d],
        ])


@from_dense
def hypercube(ndim):
    return _hypercube(ndim)


@from_row_col
def star(n):
    return np.zeros(n), np.arange(1, n + 1)


@from_dense
def lollipop(n, m):
    e = np.zeros((n + m, n + m))
    e[:n, :n] = 1 - np.eye(n)
    e[n, -m - 1] = e[n - 1, -m] = 1  # bridge
    for i in range(m, n + m):
        e[i - 1, i] = 1
    return np.triu(e)


def geometric(n_nodes, n_edges):
    n_pairs = n_nodes * (n_nodes - 1) // 2
    if n_edges > n_pairs:
        raise ValueError(
            f"There are only {n_pairs} (< {n_edges}) possible edges between " 
            f"{n_nodes} points")
    xy = np.random.random((n_nodes, 2))
    xx = row_norms(xy, squared=True)[:, np.newaxis]
    distances = np.dot(xy, xy.T)
    distances *= -2
    distances += xx
    distances += xx.T
    ur = np.triu_indices(n_nodes, k=1)
    # skip zeros and repetitions
    dist_threshold = np.partition(distances[ur], n_edges)[n_edges]
    mask = distances <= dist_threshold
    mask[np.tril_indices(n_nodes)] = False
    row, col = mask.nonzero()
    edges = sp.csr_matrix((np.ones(len(row)), (row, col)),
                          shape=(n_nodes, n_nodes))
    return Network(
        range(n_nodes), edges,
        name=f"geometric({n_nodes},{n_edges})",
        coordinates=xy
    )



import inspect
from collections import namedtuple

import numpy as np
from scipy.sparse import csgraph

from AnyQt.QtCore import QThread, Qt
from AnyQt.QtWidgets import QWidget, QGridLayout

from Orange.data import ContinuousVariable, Table, Domain
from Orange.widgets import gui, widget
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Input, Output, Msg
from orangecontrib.network.network import Network

NODELEVEL, GRAPHLEVEL, INTERNAL = range(3)
UNDIRECTED, DIRECTED, GENERAL = range(3)

ERRORED = object()
TERMINATED = object()


def shortest_paths_nan_diag(network):
    paths = csgraph.floyd_warshall(
        network.edges[0].edges, network.edges[0].directed).astype(float)
    diag = np.lib.stride_tricks.as_strided(
        paths, (len(paths), ), ((len(paths) + 1) * paths.dtype.itemsize,))
    diag[:] = np.nan
    return paths


def density(network):
    n = network.number_of_nodes()
    if n < 2:
        return 0.0

    num = network.number_of_edges()
    if not network.edges[0].directed:
        num *= 2
    return num / (n * (n - 1))


def avg_degree(network):
    m = network.number_of_edges()
    n = max(network.number_of_nodes(), 1)
    return m / n if network.edges[0].directed else 2 * m / n

# TODO: adapt betweenness_centrality and perhaps other statistics from
# https://github.com/networkdynamics/zenlib/blob/master/src/zen/algorithms/centrality.pyx

# TODO: Combo that lets the user set edge types to use (or all)

# Order of this list also defines execution priority
METHODS = (
    ("n", lambda network: network.number_of_nodes()),
    ("e", lambda network: network.number_of_edges()),
    ("degrees", lambda network: network.degrees(), "Degree", NODELEVEL, GENERAL),

    ("number_of_nodes", lambda n: n, "Number of nodes", GRAPHLEVEL, GENERAL),
    ("number_of_edges", lambda e: e, "Number of edges", GRAPHLEVEL, GENERAL),
    ("average_degree", lambda network: avg_degree(network), "Average degree", GRAPHLEVEL, GENERAL),
    ("density", lambda network: density(network), "Density", GRAPHLEVEL, GENERAL),

    ("shortest_paths", shortest_paths_nan_diag, "Shortest paths"),
    ("diameter",
     lambda shortest_paths: np.nanmax(shortest_paths), "Diameter", GRAPHLEVEL, GENERAL),
    ("radius",
     lambda shortest_paths: np.nanmin(np.nanmax(shortest_paths, axis=1)),
     "Radius", GRAPHLEVEL, GENERAL),
    ("average_shortest_path_length",
     lambda shortest_paths: np.nanmean(shortest_paths),
     "Average shortest path length", GRAPHLEVEL, GENERAL),

    ("number_strongly_connected_components", lambda network:
         csgraph.connected_components(network.edges[0].edges, False)[0],
         "Number of strongly connected components", GRAPHLEVEL, DIRECTED),
    ("number_weakly_connected_components", lambda network:
         csgraph.connected_components(network.edges[0].edges, True)[0],
         "Number of weakly connected components", GRAPHLEVEL, DIRECTED),

    ("in_degrees", lambda network: network.in_degrees(), "In-degree", NODELEVEL, GENERAL),
    ("out_degrees", lambda network: network.out_degrees(), "Out-degree", NODELEVEL, GENERAL),
    ("average_neighbour_degrees",
     lambda network, degrees: np.fromiter(
         (np.mean(degrees[network.neighbours(i)]) if degree else np.nan
          for i, degree in enumerate(degrees)),
         dtype=float, count=len(degrees)),
     "Average neighbor degree", NODELEVEL, GENERAL),
    ("degree_centrality",
     lambda degrees, n: n and degrees / (n - 1) if n > 1 else 0,
     "Degree centrality", NODELEVEL, GENERAL),
    ("in_degree_centrality",
     lambda in_degrees, n: in_degrees / (n - 1) if n > 1 else 0,
     "In-degree centrality", NODELEVEL, GENERAL),
    ("out_degree_centrality",
     lambda out_degrees, n: out_degrees / (n - 1) if n > 1 else 0,
     "Out-degree centrality", NODELEVEL, GENERAL),
    ("closeness_centrality",
     lambda shortest_paths: 1 / np.nanmean(shortest_paths, axis=1),
     "Closeness centrality", NODELEVEL, GENERAL)
)

MethodDefinition = namedtuple(
    "MethodDefinition",
    ["name", "func", "label", "level", "edge_constraint", "args"])


METHODS = {definition[0]:
    MethodDefinition(*(definition + ("", None, "", INTERNAL, GENERAL)[len(definition):]),
                     inspect.getfullargspec(definition[1]).args)
    for definition in METHODS}

"""
        ("degree_assortativity_coefficient", False, \
            "Degree assortativity coefficient", GRAPHLEVEL, \
                nx.degree_assortativity_coefficient if \
                hasattr(nx, "degree_assortativity_coefficient") else None),
        ("degree_pearson_correlation_coefficient", False, \
            "Degree pearson correlation coefficient", GRAPHLEVEL, \
            nx.degree_pearson_correlation_coefficient if\
            hasattr(nx, "degree_pearson_correlation_coefficient") else None),
        ("estrada_index", False, "Estrada index", GRAPHLEVEL, \
            nx.estrada_index if hasattr(nx, "estrada_index") else None),
        ("graph_clique_number", False, "Graph clique number", GRAPHLEVEL, nx.graph_clique_number),
        ("graph_number_of_cliques", False, "Graph number of cliques", GRAPHLEVEL, nx.graph_number_of_cliques),
        ("transitivity", False, "Graph transitivity", GRAPHLEVEL, nx.transitivity),
        ("average_clustering", False, "Average clustering coefficient", GRAPHLEVEL, nx.average_clustering),
        ("number_attracting_components", False, "Number of attracting components", GRAPHLEVEL, nx.number_attracting_components),

        ("clustering", False, "Clustering coefficient", NODELEVEL, nx.clustering),
        ("triangles", False, "Number of triangles", NODELEVEL, nx.triangles),
        ("square_clustering", False, "Squares clustering coefficient", NODELEVEL, nx.square_clustering),
        ("number_of_cliques", False, "Number of cliques", NODELEVEL, nx.number_of_cliques),
        ("betweenness_centrality", False, "Betweenness centrality", NODELEVEL, nx.betweenness_centrality),
        ("current_flow_closeness_centrality", False, "Information centrality", NODELEVEL, nx.current_flow_closeness_centrality),
        ("current_flow_betweenness_centrality", False, "Random-walk betweenness centrality", NODELEVEL, nx.current_flow_betweenness_centrality),
        ("approximate_current_flow_betweenness_centrality", False, \
            "Approx. random-walk betweenness centrality", NODELEVEL, \
            nx.approximate_current_flow_betweenness_centrality if \
            hasattr(nx, "approximate_current_flow_betweenness_centrality") \
                else None),
        ("eigenvector_centrality", False, "Eigenvector centrality", NODELEVEL, nx.eigenvector_centrality),
        ("eigenvector_centrality_numpy", False, "Eigenvector centrality (NumPy)", NODELEVEL, nx.eigenvector_centrality_numpy),
        ("load_centrality", False, "Load centrality", NODELEVEL, nx.load_centrality),
        ("core_number", False, "Core number", NODELEVEL, nx.core_number),
        ("eccentricity", False, "Eccentricity", NODELEVEL, nx.eccentricity),
        ("closeness_vitality", False, "Closeness vitality", NODELEVEL, nx.closeness_vitality),
        """


class WorkerThread(QThread):
    def __init__(self, method, data):
        super().__init__()
        self.method = method
        self.data = data

        self.stopped = 0
        self.result = None
        self.error = None
        self.is_terminated = False

    def run(self):
        args = tuple(self.data[arg] for arg in self.method.args)
        if any(arg is TERMINATED for arg in args):  # "in" doesn't work with np
            self.result = TERMINATED
        elif any(arg is ERRORED for arg in args):
            self.result = TERMINATED
        else:
            try:
                self.result = self.method.func(*args)
            except Exception as ex:
                self.error = ex
                print(ex)


class OWNxAnalysis(widget.OWWidget):
    name = 'Network Analysis'
    description = 'Statistical analysis of network data.'
    icon = 'icons/NetworkAnalysis.svg'
    priority = 6425

    resizing_enabled = False

    class Inputs:
        network = Input("Network", Network)
        items = Input("Items", Table)

    class Outputs:
        network = Output("Network", Network)
        items = Output("Items", Table)

    class Information(widget.OWWidget.Information):
        computing = Msg("Computing {}")

    want_main_area = False
    want_control_area = True

    auto_commit = Setting(False)
    enabled_methods = Setting(
        {"number_of_nodes", "number_of_edges", "average_degree"})

    def __init__(self):
        super().__init__()

        self.graph = None
        self.items = None          # items set by Items signal
        self.items_graph = None    # items set by graph.items by Network signal
        self.items_analysis = None # items to output and merge with analysis result

        self.known = {}
        self.running_jobs = {}
        # Indicates that node level statistics have changed or are pending to
        self._nodelevel_invalidated = False

        self.controlArea = QWidget(self.controlArea)
        self.layout().addWidget(self.controlArea)
        layout = QGridLayout()
        self.controlArea.setLayout(layout)
        layout.setContentsMargins(4, 4, 4, 4)
        tabs = gui.tabWidget(self.controlArea)
        tabs.setMinimumWidth(450)
        graph_indices = gui.createTabPage(tabs, "Graph-level indices",
                                          orientation=Qt.Horizontal)
        node_indices = gui.createTabPage(tabs, "Node-level indices",
                                         orientation=Qt.Horizontal)

        graph_methods = gui.vBox(graph_indices)
        gui.rubber(graph_indices)
        graph_labels = gui.vBox(graph_indices)

        node_methods = gui.vBox(node_indices)
        gui.rubber(node_indices)
        node_labels = gui.vBox(node_indices)
        graph_labels.layout().setAlignment(Qt.AlignRight)

        self.method_cbs = {}
        for method in METHODS.values():
            if method.level == INTERNAL:
                continue
            setattr(self, method.name, method.name in self.enabled_methods)
            setattr(self, "lbl_" + method.name, "")

            methods = node_methods if method.level == NODELEVEL else graph_methods
            labels = node_labels if method.level == NODELEVEL else graph_labels

            cb = gui.checkBox(
                methods, self, method.name, method.label,
                callback=lambda attr=method.name: self.method_clicked(attr)
            )
            self.method_cbs[method.name] = cb

            lbl = gui.label(labels, self, f"%(lbl_{method.name})s")
            labels.layout().setAlignment(lbl, Qt.AlignRight)
            setattr(self, "tool_" + method.name, lbl)
            # todo: is this accessible through controls?
        graph_indices.layout().addStretch(1)
        node_indices.layout().addStretch(1)

    @Inputs.network
    def set_graph(self, graph):
        self.cancel_job()
        self.graph = graph

        allowed_edge_types = {GENERAL, DIRECTED, UNDIRECTED}
        if graph is not None:
            allowed_edge_types.remove(UNDIRECTED if self.graph.edges[0].directed else DIRECTED)

        self.known = {}
        for name in METHODS:
            curr_method = METHODS[name]
            if curr_method.level == INTERNAL:
                continue

            lbl_obj = getattr(self, "tool_{}".format(name))
            cb_obj = self.method_cbs[name]

            # disable/re-enable valid graph indices
            if curr_method.edge_constraint not in allowed_edge_types:
                lbl_obj.setDisabled(True)
                cb_obj.setChecked(False)
                cb_obj.setEnabled(False)
            else:
                lbl_obj.setDisabled(False)
                cb_obj.setEnabled(True)

            setattr(self, f"lbl_{name}", "")

        if graph is not None:
            self.known["network"] = graph
            self.items_graph = graph.nodes

    @Inputs.items
    def set_items(self, items):
        self.items = items

    def handleNewSignals(self):
        if self.items is not None:
            self.items_analysis = self.items
        elif self.graph:
            self.items_analysis = self.graph.nodes
        else:
            self.items_analysis = None
        self._nodelevel_invalidated = True
        self.run_more_jobs()

    def needed_methods(self):
        # Preconditions for methods could be precomputed, so this function would
        # only compute the union of sets of conditions for enabled checkboxes
        tasks = [
            name for name in METHODS if getattr(self, name, False)]
        for name in tasks:
            tasks += [name for name in METHODS[name].args if name != "network"]
        tasks = set(tasks) - set(self.known)
        return [method for name, method in METHODS.items() if name in tasks]

    def run_more_jobs(self):
        known = set(self.known)
        needed = self.needed_methods()
        for method in needed:
            if method.name not in self.running_jobs:
                setattr(self, "lbl_" + method.name, "pending")
        doable = [method for method in needed
                  if method.name not in self.running_jobs
                  and set(method.args) <= known]
        free = max(1, QThread.idealThreadCount()) - len(self.running_jobs)
        if not doable:
            # This will output new data when everything is finished
            self.send_data()
        for method in doable[:free]:
            job = WorkerThread(method, self.known)
            job.finished.connect(lambda job=job: self.job_finished(job))
            self.running_jobs[method.name] = job
            job.start()
            if not method.level == INTERNAL:
                setattr(self, "lbl_" + method.name, "running")
        self.show_computing()

    def job_finished(self, job):
        method = job.method
        self.known[method.name] = job.result
        del self.running_jobs[method.name]
        self.set_label_for(method.name)
        self.run_more_jobs()

    def set_label_for(self, name):
        level = METHODS[name].level
        if level == INTERNAL:
            return
        value = self.known.get(name, None)
        txt = ""
        if getattr(self, name, False):
            if value is TERMINATED:
                txt = "terminated"
            elif value is ERRORED:
                txt = "error"
            elif value is None:
                txt = "computing" if name in self.running_jobs else "pending"
            elif level == GRAPHLEVEL:
                txt = f"{value:.4g}"
        setattr(self, "lbl_" + name, txt)

    def show_computing(self):
        computing = ", ".join(METHODS[name].label for name in self.running_jobs)
        self.Information.computing(computing, shown=bool(computing))

    def cancel_job(self, name=None):
        # This does not really work because functions called in those
        # threads do not observe the "is_terminated" flag and won't quit
        if name is None:
            to_stop = list(self.running_jobs)
        elif name in self.running_jobs:
            to_stop = [name]
        else:
            # This task is not running; but are its preconditions running?
            still_needed = self.needed_methods()
            to_stop = [
                name for name in (job.method.name for job in self.running_jobs)
                if METHODS[name] not in still_needed]
        for name in to_stop:
            job = self.running_jobs[name]
            job.is_terminated = True
            job.finished.disconnect()
            job.quit()
        for name in to_stop:
            job = self.running_jobs[name]
            job.wait()
            setattr(self, "lbl_" + name, "terminated")
            del self.running_jobs[name]
        self.show_computing()

    def onDeleteWidget(self):
        self.cancel_job()
        super().onDeleteWidget()

    def send_data(self):
        # Don't send when computation is still on, or it's done but no node
        # level statistics have changed
        if self.running_jobs or not self._nodelevel_invalidated:
            return
        self._nodelevel_invalidated = False
        if self.graph is None:
            self.Outputs.network.send(None)
            self.Outputs.items.send(None)
            return

        to_report = [
            method for attr, method in METHODS.items()
            if method.level == NODELEVEL
            and getattr(self, attr) and attr in self.known]
        items = self.items_analysis
        graph = self.graph
        n = graph.number_of_nodes()
        if isinstance(items, Table):
            dom = self.items_analysis.domain
            attrs, class_vars, metas = dom.attributes, dom.class_vars, dom.metas
            x, y, m = items.X, items.Y, items.metas
        else:
            attrs, class_vars, metas = [], [], []
            x = y = m = np.empty((n, 0))
        attrs += tuple(ContinuousVariable(method.label) for method in to_report)
        x = np.hstack(
            (x, ) + tuple(self.known[method.name].reshape((n, 1))
                          for method in to_report))
        domain = Domain(attrs, class_vars, metas)
        table = Table(domain, x, y, m)
        new_graph = Network(table, graph.edges, graph.name, graph.coordinates)
        self.Outputs.network.send(new_graph)
        self.Outputs.items.send(table)

    def method_clicked(self, name):
        if METHODS[name].level == NODELEVEL:
            self._nodelevel_invalidated = True
        if getattr(self, name):
            self.enabled_methods.add(name)
            if name in self.known:
                self.set_label_for(name)
                self.send_data()
            else:
                self.run_more_jobs()
        else:
            self.enabled_methods.remove(name)
            if name in self.running_jobs:
                self.cancel_job(name)
            else:
                self.set_label_for(name)
            self.send_data()

    def send_report(self):
        self.report_items("", items=[(method.label, f"{self.known[attr]:.4g}")
                                     for attr, method in METHODS.items()
                                     if method.level == GRAPHLEVEL
                                     and getattr(self, attr)
                                     and attr in self.known
                                     and isinstance(self.known[attr], (int, float))])


def main():
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    from orangecontrib.network.network.readwrite \
        import read_pajek, transform_data_to_orange_table
    from os.path import join, dirname

    #network = read_pajek(join(dirname(dirname(__file__)), 'networks', 'leu_by_genesets.net'))
    network = read_pajek(join(dirname(dirname(__file__)), 'networks', 'lastfm.net'))
    #network = read_pajek(join(dirname(dirname(__file__)), 'networks', 'Erdos02.net'))
    #transform_data_to_orange_table(network)
    WidgetPreview(OWNxAnalysis).run(set_graph=network)


if __name__ == "__main__":
    main()



"""
=======
Network
=======

Network visualization and analysis tools for Orange.

"""

# Category description for the widget registry
import sysconfig
from orangecontrib.network import Network
from orangewidget.utils.signals import summarize, PartialSummary

NAME = "Network"

DESCRIPTION = "Network visualization and analysis tools for Orange."

BACKGROUND = "#C0FF97"

ICON = "icons/Category-Network.svg"

# Location of widget help files.
WIDGET_HELP_PATH = (
    # Used for development.
    # You still need to build help pages using
    # make htmlhelp
    # inside doc folder
    ("{DEVELOP_ROOT}/doc/_build/html/index.html", None),

    # Documentation included in wheel
    # Correct DATA_FILES entry is needed in setup.py and documentation has to be built
    # before the wheel is created.
    ("{}/help/orange3-network/index.html".format(sysconfig.get_path("data")), None),

    # Online documentation url, used when the local documentation is available.
    # Url should point to a page with a section Widgets. This section should
    # includes links to documentation pages of each widget. Matching is
    # performed by comparing link caption to widget name.
    ("https://orange3-network.readthedocs.io/en/latest/", "")
)


@summarize.register
def summarize_(net: Network):
    n = net.number_of_nodes()
    e = net.number_of_edges()
    if len(net.edges) == 1:
        directed = net.edges[0].directed
        direct = "➝" if directed else "–"
        nettype = ['Network', 'Directed network'][directed]
        details = f"<nobr>{nettype} with {n} nodes " \
                  f"and {net.number_of_edges()} edges</nobr>."
    else:
        direct = "–"
        details = f"<nobr>Network with {n} nodes"
        if net.edges:
            details += " and {len(net.edges)} edge types:</nobr><ul>" + "".join(
                f"<li>{len(edges)} edges, "
                f"{['undirected', 'directed'][edges.directed]}</li>"
                for edges in net.edges) + "."

    return PartialSummary(f"•{n} {direct}{e}", details)



from orangecontrib.network.widgets.OWNxGenerator import OWNxGenerator
from orangecontrib.network.widgets.tests.utils import NetworkTest


class TestOWNxGenerator(NetworkTest):

    def setUp(self):
        self.widget = self.create_widget(OWNxGenerator)  # type: OWNxGenerator



import os
import unittest
from unittest.mock import patch, Mock

import numpy as np

import Orange
from orangecontrib.network.widgets.OWNxFile import OWNxFile
from orangecontrib.network.widgets.tests.utils import NetworkTest

TEST_NETS = os.path.join(os.path.split(__file__)[0], "networks")

def _get_test_net(filename):
    return os.path.join(TEST_NETS, filename)


class TestOWNxFile(NetworkTest):
    def setUp(self):
        self.widget = self.create_widget(OWNxFile)  # type: OWNxFile

    def test_read_error(self):
        with patch("orangecontrib.network.widgets.OWNxFile.read_pajek",
                   Mock(side_effect=OSError)):
            self.widget.open_net_file("foo.net")
        self.assertTrue(self.widget.Error.io_error.is_shown())
        filename = self._get_filename("leu_by_genesets.net")
        self.widget.open_net_file(filename)
        self.assertFalse(self.widget.Error.io_error.is_shown())

    def test_load_datafile(self):
        self.widget.open_net_file(_get_test_net("test.net"))
        items = self.get_output(self.widget.Outputs.items)
        self.assertEqual(items[0]["name"], "aaa")

    def test_invalid_datafile_length(self):
        # When data file's length does not match, the widget must create
        # a table from node labels
        self.widget.open_net_file(_get_test_net("test_inv.net"))
        self.assertTrue(self.widget.Error.mismatched_lengths)

        network = self.get_output(self.widget.Outputs.network)
        self.assertEqual(network.number_of_nodes(), 7)

        items = self.get_output(self.widget.Outputs.items)
        self.assertEqual(len(items), 7)
        self.assertEqual(items[0]["node_label"], "aa")

    def test_vars_for_label(self):
        self.widget.open_net_file(self._get_filename(None, mode="t"))

        data = Orange.data.Table(_get_test_net("test_data.tab"))
        domain = data.domain
        best_var, useful_vars = self.widget._vars_for_label(data)
        self.assertIs(best_var, domain["label"])
        self.assertEqual(useful_vars, [domain["with_extras"], domain["label"]])

        data = Orange.data.Table.from_list(
            Orange.data.Domain(
                [], None, [Orange.data.StringVariable(x) for x in "abcde"]),
            [["aa", "", "cc", "aa", ""],
             ["bb", "bb", "cc", "bb", "aa"],
             ["cc", "", "aa", "cc", "bb"],
             ["dd", "aa", "bb", "dd", "cc"],
             ["ee", "cc", "dd", "ee", ""],
             ["ff", "ee", "ee", "ff", "dd"],
             ["gg", "dd", "ff", "", "ee"],
             ["hh", "ff", "gg", "", "ff"],
             ["ii", "gg", "", "", "gg"]]
        )
        domain = data.domain
        best_var, useful_vars = self.widget._vars_for_label(data)
        self.assertIs(best_var, domain["b"])
        # c is not unique and d doesn't cover all values
        self.assertEqual(useful_vars, [domain["a"], domain["b"], domain["e"]])

    def test_label_combo_contents(self):
        widget = self.widget
        widget.read_auto_data = Mock()

        widget.open_net_file(self._get_filename(None, mode="t"))
        self.assertEqual(list(widget.label_model), [None])

        data = Orange.data.Table(_get_test_net("test_data.tab"))
        domain = data.domain

        self.send_signal(widget.Inputs.items, data)

        # Model contains useful variables
        self.assertEqual(list(widget.label_model),
                         [None, domain["with_extras"], domain["label"]])

        # `label` is chosen as default, and output has corresponding data
        self.assertIs(widget.label_variable, domain["label"])
        output = self.get_output(widget.Outputs.network)
        id_col = output.nodes.get_column("id")
        np.testing.assert_equal(id_col, np.arange(1, 8))

        # No variable, row matching. Error is shown and original labels are used
        widget._label_to_tabel = Mock(return_value=data[:7])
        widget.label_variable = None
        widget.label_changed()
        self.assertTrue(widget.Error.mismatched_lengths.is_shown())
        output = self.get_output(widget.Outputs.network)
        self.assertIs(output.nodes, widget._label_to_tabel.return_value)

        # Choose a different variable; no error, output has corresponding data
        widget.label_variable = domain["with_extras"]
        widget.label_changed()
        self.assertFalse(widget.Error.mismatched_lengths.is_shown())
        output = self.get_output(widget.Outputs.network)
        id_col = output.nodes.get_column("id")
        np.testing.assert_equal(id_col, np.arange(2, 9))

        # Remove data: model must be cleared, data back to original
        self.send_signal(widget.Inputs.items, None)
        self.assertEqual(list(widget.label_model), [None])
        output = self.get_output(widget.Outputs.network)
        self.assertIs(output.nodes, widget._label_to_tabel.return_value)

        # Bring data back, and turn on row matching;
        # this triggers an error; then remove data; error must be gone
        self.send_signal(widget.Inputs.items, data)
        widget.label_variable = None
        widget.label_changed()
        self.assertTrue(widget.Error.mismatched_lengths.is_shown())

        self.send_signal(widget.Inputs.items, None)
        output = self.get_output(widget.Outputs.network)
        self.assertIs(output.nodes, widget._label_to_tabel.return_value)
        self.assertFalse(widget.Error.mismatched_lengths.is_shown())

    def test_context_matching(self):
        widget = self.widget
        widget.open_net_file(self._get_filename(None, mode="t"))
        data = Orange.data.Table(_get_test_net("test_data.tab"))
        domain = data.domain

        self.send_signal(widget.Inputs.items, data)
        self.assertIs(widget.label_variable, domain["label"])

        widget.label_variable = domain["with_extras"]
        self.send_signal(widget.Inputs.items, None)
        self.assertIs(widget.label_variable, None)

        self.send_signal(widget.Inputs.items, data)
        self.assertIs(widget.label_variable, domain["with_extras"])


if __name__ == "__main__":
    unittest.main()



# pylint: disable=protected-access

import unittest

import numpy as np
import scipy.sparse as sp
import orangecontrib.network.network.twomode as tm
from orangecontrib.network import Network
from orangecontrib.network.widgets.tests.utils import _create_net


class TestTwoMode(unittest.TestCase):
    def test_normalization(self):
        def edges_almost_equal(a, edges):
            a = a.todense()
            a2 = np.zeros(a.shape)
            for n1, n2, w in edges:
                a2[n1, n2] = w
            np.testing.assert_almost_equal(a, a2)
            self.assertEqual(a.dtype, float)

        edges = sp.coo_matrix(
            ([1., 5, 3, 4, 2, 6], ([0, 1, 1, 2, 2, 3], [4, 4, 5, 4, 5, 6])),
            (7, 7)
        )

        s = np.sqrt
        edges_almost_equal(
            tm.Weighting[tm.NoWeights].func(edges, 3),
            ((0, 1, 1), (0, 2, 1), (1, 2, 1))
        )
        edges_almost_equal(
            tm.Weighting[tm.WeightConnections].func(edges, 3),
            ((0, 1, 1), (0, 2, 1), (1, 2, 2))
        )
        edges_almost_equal(
            tm.Weighting[tm.WeightWeightedConnections].func(edges, 3),
            ((0, 1, 5), (0, 2, 4), (1, 2, 26))
        )
        edges_almost_equal(
            tm.Weighting[tm.WeightGeo].func(edges, 3),
            ((0, 1, 1 / s(10) * 5 / s(80)),
             (0, 2, 1 / s(10) * 4 / s(60)),
             (1, 2, 5 / s(80) * 4 / s(60) + 3 / s(40) * 2 / s(30)))
        )
        edges_almost_equal(
            tm.Weighting[tm.WeightGeoDeg].func(edges, 3),
            ((0, 1, 1 / s(3) * 5 / s(6)),
             (0, 2, 1 / s(3) * 4 / s(6)),
             (1, 2, 5 / s(6) * 4 / s(6) + 3 / s(4) * 2 / s(4)))
        )
        edges_almost_equal(
            tm.Weighting[tm.WeightInput].func(edges, 3),
            ((0, 1, 1 / 1 * 5 / 8),
             (0, 2, 1 / 1 * 4 / 6),
             (1, 2, 5 / 8 * 4 / 6 + 3 / 8 * 2 / 6))
        )
        edges_almost_equal(
            tm.Weighting[tm.WeightOutput].func(edges, 3),
            ((0, 1, 1 / 10 * 5 / 10),
             (0, 2, 1 / 10 * 4 / 10),
             (1, 2, 5 / 10 * 4 / 10 + 3 / 5 * 2 / 5))
        )
        edges_almost_equal(
            tm.Weighting[tm.WeightMin].func(edges, 3),
            ((0, 1, 1 / 1 * 5 / 8),
             (0, 2, 1 / 1 * 4 / 6),
             (1, 2, 5 / 8 * 4 / 6 + 3 / 5 * 2 / 5))
        )
        edges_almost_equal(
            tm.Weighting[tm.WeightMax].func(edges, 3),
            ((0, 1, 1 / 10 * 5 / 10),
             (0, 2, 1 / 10 * 4 / 10),
             (1, 2, 5 / 10 * 4 / 10 + 3 / 8 * 2 / 6))
        )

    def test_filtered_edges(self):
        def assert_edges(actual, expected):
            self.assertEqual(len(actual.data), len(expected))
            self.assertEqual(actual.data.dtype, float)
            self.assertEqual(
                set(zip(actual.row, actual.col, actual.data)), set(expected))

        net = _create_net(((0, 4, 1.), (4, 1, 5), (1, 5, 3),
                           (2, 4, 4), (2, 5, 2), (3, 6, 6)))

        # All edges
        assert_edges(
            tm._filtered_edges(
                net,
                np.array([True] * 4 + [False] * 3),
                np.array([False] * 4 + [True] * 3)),
            ((0, 4, 1.), (1, 4, 5.), (1, 5, 3.),
             (2, 4, 4.), (2, 5, 2.), (3, 6, 6.))
        )

        # All edges, opposite mode roles
        assert_edges(
            tm._filtered_edges(
                net,
                np.array([False] * 4 + [True] * 3),
                np.array([True] * 4 + [False] * 3)),
            ((0, 0, 1.), (0, 1, 5.), (1, 1, 3.),
             (0, 2, 4.), (1, 2, 2.), (2, 3, 6.))
        )

        # Not all edges
        assert_edges(
            tm._filtered_edges(
                net,
                np.array([True] * 4 + [False] * 3),
                np.array([False] * 5 + [True] * 2)),
            ((1, 5, 3.), (2, 5, 2.), (3, 6, 6.))
        )

        # One mode is empty
        assert_edges(
            tm._filtered_edges(
                net,
                np.array([True] * 4 + [False] * 3),
                np.array([False] * 7)),
            ()
        )

        # The other mode is empty
        assert_edges(
            tm._filtered_edges(
                net,
                np.array([False] * 7),
                np.array([False] * 5 + [True] * 2)),
            ()
        )

        # Both modes are empty
        assert_edges(
            tm._filtered_edges(
                net,
                np.array([False] * 7),
                np.array([False] * 7)),
            ()
        )

        # Graph is empty
        net = Network(range(7), sp.csr_matrix((7, 7)))
        assert_edges(
            tm._filtered_edges(
                net,
                np.array([True] * 4 + [False] * 3),
                np.array([False] * 7)),
            ()
        )

    def test_to_single_mode(self):
        net = _create_net(((0, 4, 1.), (4, 1, 5), (1, 5, 3),
                           (2, 4, 4), (2, 5, 2), (3, 6, 6)))

        net1 = tm.to_single_mode(
            net,
            np.array([True] * 4 + [False] * 3),
            np.array([False] * 4 + [True] * 3),
            tm.NoWeights)

        np.testing.assert_equal(
            net1.edges[0].edges.todense(),
            np.array([[0, 1, 1, 0],
                      [0, 0, 1, 0],
                      [0, 0, 0, 0],
                      [0, 0, 0, 0]])
        )

        net = _create_net(((0, 1, 1.0), (0, 2, 1.0), (1, 2, 1.0), (2, 3, 1.0)), n=5)
        net2 = tm.to_single_mode(
            net,
            np.array([True, False, False, True, False]),
            np.array([False, False, True, False, True]),
            tm.NoWeights
        )
        np.testing.assert_equal(
            net2.edges[0].edges.todense(),
            np.array([[0, 1],
                      [0, 0]])
        )


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from tempfile import NamedTemporaryFile

import numpy as np

from orangecontrib.network.network import readwrite


def _fullpath(name):
    return os.path.join(os.path.split(__file__)[0], name)


class TestReadPajek(unittest.TestCase):
    def test_two_mode(self):
        davis = readwrite.read_pajek(_fullpath("../networks/davis.net"))
        self.assertEqual(davis.number_of_nodes(), 32)
        self.assertEqual(
            list(davis.nodes),
            ['EVELYN', 'LAURA', 'THERESA', 'BRENDA', 'CHARLOTTE', 'FRANCES',
             'ELEANOR', 'PEARL', 'RUTH', 'VERNE', 'MYRNA', 'KATHERINE',
             'SYLVIA', 'NORA', 'HELEN', 'DOROTHY', 'OLIVIA', 'FLORA', 'E1',
             'E2', 'E3', 'E4', 'E5', 'E6', 'E7', 'E8', 'E9', 'E10', 'E11',
             'E12', 'E13', 'E14']
            )
        self.assertEqual(davis.in_first_mode, 18)

    def test_write_pajek(self):
        net = readwrite.read_pajek(_fullpath("../networks/leu_by_genesets.net"))
        with NamedTemporaryFile("wt", suffix=".net", delete=False) as f:
            try:
                readwrite.write_pajek(f, net)
                f.close()
                net2 = readwrite.read_pajek(f.name)
                np.testing.assert_equal(net2.nodes, net.nodes)
                np.testing.assert_equal(net2.coordinates, net.coordinates)
                self.assertEqual(len(net2.edges), 1)
                edges, edges2 = net.edges[0].edges, net2.edges[0].edges
                np.testing.assert_equal(edges.indptr, edges2.indptr)
                np.testing.assert_equal(edges.indices, edges2.indices)
                np.testing.assert_almost_equal(edges.data, edges2.data)
            finally:
                os.remove(f.name)

    def test_write_pajek_no_coordinates(self):
        net = readwrite.read_pajek(_fullpath("../networks/leu_by_genesets.net"))
        net.coordinates = None
        with NamedTemporaryFile("wt", suffix=".net", delete=False) as f:
            try:
                readwrite.write_pajek(f, net)
                f.close()
                net2 = readwrite.read_pajek(f.name)
                np.testing.assert_equal(net2.nodes, net.nodes)
                self.assertIsNone(net2.coordinates, net.coordinates)
            finally:
                os.remove(f.name)

    def test_write_pajek_multiple_edge_types(self):
        net = readwrite.read_pajek(_fullpath("../networks/leu_by_genesets.net"))
        net.edges.append(net.edges[0])
        with NamedTemporaryFile("wt", suffix=".net") as f:
            self.assertRaises(TypeError, readwrite.write_pajek, f, net)

    def test_edge_list(self):
        net = readwrite.read_pajek(_fullpath("test-arcslist.net"))
        neighs = [(1, (2, 3, 6)),
                  (2, (1, 4, 5, 6)),
                  (5, (1, 2)),
                  (6, (2, 3, 4))]
        self.assertEqual(net.number_of_edges(), sum(len(y) for _, y in neighs))
        self.assertTrue(net.edges[0].directed)
        for x, y in neighs:
            np.testing.assert_equal(net.outgoing(x - 1), np.array(y) - 1)


if __name__ == "__main__":
    unittest.main()



import string
from collections import defaultdict

import numpy as np

from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QSpinBox

from Orange.data import Table, Domain, StringVariable
from Orange.widgets import gui, widget, settings
from Orange.widgets.widget import Output, Msg

from orangecontrib.network import Network
# __all__ is defined, pylint: disable=wildcard-import, unused-wildcard-import
from orangecontrib.network.network.generate import *


class GraphType:
    """
    BALANCED_TREE = ('Balanced tree', _balanced_tree)
    REGULAR = ('Regular', lambda n: nx.random_regular_graph(min(np.random.randint(10)*2, n - 1), n))
    SCALEFREE = ('Scale-free', lambda n: nx.scale_free_graph(int(n)))
    SHELL = ('Shell', lambda n: nx.random_shell_graph([(int(n*.1), int(n*.1), .2),
                                                       (int(n*.3), int(n*.3), .8),
                                                       (int(n*.6), int(n*.6), .5)]))
    WHEEL = ('Wheel', lambda n: nx.wheel_graph(int(n)))
"""


def _ctrl_name(name, arg):
    return (name + "___" + arg).replace(" ", "_")


class OWNxGenerator(widget.OWWidget):
    name = "Network Generator"
    description = "Construct example graphs."
    icon = "icons/NetworkGenerator.svg"
    priority = 6420

    GRAPH_TYPES = (
        (path, "Path",
         "of length", (10, 2, 100)),
        (cycle, "Cycle",
         "with", (10, 3, 100), "nodes"),
        (complete, "Complete",
         "with", (5, 1, 100), "nodes"),
        (complete_bipartite, "Complete bipartite",
         "with", (5, 1, 100), "and", (8, 1, 100), "nodes"),
        (barbell, "Barbell",
         "with", (5, 1, 100), "and", (8, 1, 100), "nodes"),
        (ladder, "Ladder",
         "with", (10, 2, 100), "steps"),
        (circular_ladder, "Circular ladder",
         "with", (8, 2, 100), "steps"),
        (grid, "Grid",
         ", height", (4, 2, 100), "and width", (5, 2, 100)),
        (hypercube, "Hypercube",
         ", ", (4, 1, 14), "dimensional"),
        (star, "Star", "with", (10, 1, 1000), "edges"),
        (lollipop, "Lollipop",
         "with", (5, 3, 30), "nodes, stem of", (5, 3, 100), "nodes"),
        (geometric, "Geometric",
         "with", (20, 2, 100), "nodes, ", (50, 1, 10000), "edges")
    )
    RANDOM_TYPES = ("Geometric", )

    class Outputs:
        network = Output("Network", Network)

    class Error(widget.OWWidget.Error):
        generation_error = Msg("{}")

    graph_type = settings.Setting(7)
    arguments = settings.Setting(
        {func.__name__: [args[0] for args in defaults if isinstance(args, tuple)]
         for func, _, *defaults in GRAPH_TYPES})
    settings_version = 3

    want_main_area = False
    resizing_enabled = False

    def __init__(self):
        def space(a, b):
            if not isinstance(a, str) or a[-1] in string.ascii_letters \
                    and not isinstance(b, str) or b[0] in string.ascii_letters:
                return " "
            else:
                return ""

        super().__init__()
        rb = self.radios = gui.radioButtons(
            self.controlArea, self, "graph_type",
            box=True,
            callback=self.on_type_changed
        )
        rb.layout().setSpacing(6)
        self.arg_spins = {}
        for (func, name, _, *arguments), defaults \
                in zip(self.GRAPH_TYPES, self.arguments.values()):
            argbox = gui.hBox(rb)
            argbox.layout().setSpacing(0)
            rbb = gui.appendRadioButton(rb, name, argbox)
            rbb.setAttribute(Qt.WA_LayoutUsesWidgetRect)
            argbox.setAttribute(Qt.WA_LayoutUsesWidgetRect)
            self.arg_spins[name] = box = []
            values = iter(defaults)
            argno = 0
            for prev, arg, post in \
                    zip([""] + arguments, arguments, arguments + [""]):
                if isinstance(arg, str):
                    label = space(prev, arg) + arg + space(arg, post)
                    box.append(gui.widgetLabel(argbox, label))
                else:
                    assert isinstance(arg, tuple)
                    _value, minv, maxv = arg
                    spin = QSpinBox(value=next(values), minimum=minv, maximum=maxv)
                    argbox.layout().addWidget(spin)
                    spin.valueChanged.connect(
                        lambda value, name=func.__name__, argidx=argno:
                        self.update_arg(value, name, argidx))
                    argno += 1
                    box.append(spin)

        self.bt_generate = gui.button(
            self.controlArea, self, "Regenerate Network",
            callback=self.generate)
        self.on_type_changed()

    def on_type_changed(self):
        cur_def = self.GRAPH_TYPES[self.graph_type]
        cur_name = cur_def[1]
        for (name, widgets), (_1, _2, name_add, *_), radio in \
                zip(self.arg_spins.items(), self.GRAPH_TYPES, self.radios.buttons):
            selected = name == cur_name
            radio.setText(name
                + (" " * (name_add and name_add[0] in string.ascii_letters)
                   + name_add) * selected)
            for spin in widgets:
                spin.setHidden(not selected)

        self.bt_generate.setEnabled(cur_name in self.RANDOM_TYPES)

        self.generate()

    def update_arg(self, value, name, argidx):
        self.arguments[name][argidx] = value
        self.generate()

    def generate(self):
        func, name, args, *_ = self.GRAPH_TYPES[self.graph_type]
        args = self.arguments[func.__name__]
        self.Error.generation_error.clear()
        try:
            network = func(*args)
        except ValueError as exc:
            self.Error.generation_error(exc)
            network = None
        else:
            n = len(network.nodes)
            network.nodes = Table(Domain([], [], [StringVariable("id")]),
                                  np.zeros((n, 0)), np.zeros((n, 0)),
                                  np.arange(n).reshape((n, 1)))
        self.Outputs.network.send(network)

    @classmethod
    def migrate_settings(cls, settings_, version):
        if version < 2:
            arguments = defaultdict(list)
            for name, value in settings_.pop("arguments", {}):
                arguments[name.split("__")[0]].append(value)
            settings_["arguments"] = dict(arguments)

        if version < 3:
            old = settings_["arguments"]
            settings_["arguments"] = {
                func.__name__: old.get(name, [args[0] for args in defaults
                                              if isinstance(args, tuple)])
                for func, name, *defaults in cls.GRAPH_TYPES}


def main():
    def send(graph):
        owe.set_graph(graph)
        owe.handleNewSignals()

    from AnyQt.QtWidgets import QApplication
    from orangecontrib.network.widgets.OWNxExplorer import OWNxExplorer
    a = QApplication([])
    ow = OWNxGenerator()
    owe = OWNxExplorer()
    ow.Outputs.network.send = send
    owe.show()
    ow.show()
    a.exec_()
    ow.saveSettings()


if __name__ == "__main__":
    main()



from importlib.resources import files, as_file

def networks():
    path_in_package = files(__package__).joinpath('networks')
    with as_file(path_in_package) as dir_path:
        yield ('', dir_path)

from .network import *

from orangecontrib.network.network import community



from functools import reduce, wraps, partial
from typing import Sequence

import numpy as np
import scipy.sparse as sp
from numpy.lib.stride_tricks import as_strided


class Edges:
    directed = None

    def __init__(self,
                 edges,  # row=from, column=to
                 edge_data: Sequence = None,
                 name: str = ""):
        self.edges = edges.tocsr(copy=True)
        self.edges.sum_duplicates()
        self.edge_data = edge_data
        self.name = name

    def out_degrees(self, *, weighted=False) -> np.ndarray:
        pass

    def in_degrees(self, *, weighted=False) -> np.ndarray:
        pass

    def degrees(self, *, weighted=False) -> np.ndarray:
        pass

    def out_degree(self, node, *, weighted=False) -> float:
        pass

    def in_degree(self, node, *, weighted=False) -> float:
        pass

    def degree(self, node, *, weighted=False) -> float:
        pass

    def outgoing(self, node, weights=False) -> np.ndarray:
        pass

    def ingoing(self, node, weights=False) -> np.ndarray:
        pass

    def neighbours(self, node, edge_type=None, weights=False) -> np.ndarray:
        pass

    @staticmethod
    def _compose_neighbours(node: int, matrix: sp.csr_matrix, weights: bool):
        fr, to = matrix.indptr[node], matrix.indptr[node + 1]
        if not weights:
            return matrix.indices[fr:to]
        else:
            return matrix.indices[fr:to], matrix.data[fr:to]

    @staticmethod
    def _compute_degrees(edges, weighted):
        if weighted:
            return edges.sum(axis=1).getA1()
        else:
            return edges.indptr[1:] - edges.indptr[:-1]

    @staticmethod
    def _compute_degree(edges, node, weighted):
        fr, to = edges.indptr[node], edges.indptr[node + 1]
        if weighted:
            return edges.data[fr:to].sum()
        else:
            return to - fr

    def subset(self, mask, node_renumeration, shape):
        edges = self.edges.tocoo()
        edge_mask = np.logical_and(mask[edges.row], mask[edges.col])
        row = node_renumeration[edges.row[edge_mask]]
        col = node_renumeration[edges.col[edge_mask]]
        data = edges.data[edge_mask]
        edge_data = self.edge_data[edge_mask] if self.edge_data is not None \
            else None
        return type(self)(
            sp.csr_matrix(
                (data, (row, col)), shape=shape), edge_data, self.name)


class DirectedEdges(Edges):
    directed = True

    def __init__(self,
                 edges: sp.csr_matrix,  # row=from, column=to
                 edge_data: Sequence = None,
                 name: str = ""):
        super().__init__(edges, edge_data, name)
        self.in_edges = self.edges.transpose()

    def out_degrees(self, *, weighted=False):
        return self._compute_degrees(self.edges, weighted)

    def in_degrees(self, *, weighted=False):
        return self._compute_degrees(self.in_edges, weighted)

    def degrees(self, *, weighted=False):
        return self._compute_degrees(self.edges, weighted) \
               + self._compute_degrees(self.in_edges, weighted)

    def out_degree(self, node, *, weighted=False):
        return self._compute_degree(self.edges, node, weighted)

    def in_degree(self, node, *, weighted=False):
        return self._compute_degree(self.in_edges, node, weighted)

    def degree(self, node, *, weighted=False):
        return self._compute_degree(self.in_edges, node, weighted) \
               + self._compute_degree(self.out_Edgesnode, weighted)

    def outgoing(self, node, weights=False):
        return self._compose_neighbours(node, self.edges, weights)

    def incoming(self, node, weights=False):
        return self._compose_neighbours(node, self.in_edges, weights)

    def neighbours(self, node, edge_type=None, weights=False):
        return np.hstack(
            (self._compose_neighbours(node, self.edges, weights),
             self._compose_neighbours(node, self.in_edges, weights)))


class UndirectedEdges(Edges):
    directed = False

    def __init__(self,
                 edges: sp.csr_matrix,
                 edge_data: Sequence = None,
                 name: str = ""):
        super().__init__(edges, edge_data, name)
        self.twoway_edges = self._make_twoway_edges()

    def _make_twoway_edges(self):
        edges = self.edges.copy()
        n_edges = len(edges.data)
        zero_strided = self.edges.data.strides == (0, )

        if not n_edges:
            return edges

        # Replaces 0s with max + 1 so sparse operations don't remove them
        if zero_strided:
            max_weight = edges.data[0]
            # Save (temporary) memory and CPU time
            edges.data = as_strided(1, (n_edges, ), (0,))
        else:
            max_weight = np.max(edges.data)
            edges.data[edges.data == 0] = max_weight + 1

        twe = edges + edges.transpose()
        twe.sum_duplicates()

        if zero_strided:
            # Save memory
            twe.data = as_strided(max_weight, (n_edges, ), (0,))
        else:
            twe.data[twe.data > max_weight] = 0
        return twe

    def degrees(self, *, weighted=False):
        return self._compute_degrees(self.twoway_edges, weighted)

    def degree(self, node, *, weighted=False):
        return self._compute_degree(self.twoway_edges, node, weighted)

    def neighbours(self, node, weights=False):
        return self._compose_neighbours(node, self.twoway_edges, weights)

    in_degrees = out_degrees = degrees
    in_degree = out_degree = degree
    incoming = outgoing = neighbours


EdgeType = [UndirectedEdges, DirectedEdges]


def aggregate_over_edge_types(aggregate, arg_no=0):
    def wrapwrap(f):
        @wraps(f)
        def wrapper(graph, *args, **kwargs):
            if len(args) <= arg_no or args[arg_no] is None:
                return aggregate([
                    f(graph,
                      *args[:arg_no], edge_type, *args[arg_no + 1:],
                      **kwargs)
                    for edge_type in range(len(graph.edges))])
            else:
                return f(graph, *args, **kwargs)
        return wrapper
    return wrapwrap


sum_over_edge_types = \
    partial(aggregate_over_edge_types, lambda x: reduce(np.add, x, 0))

concatenate_over_edge_types = partial(aggregate_over_edge_types, np.hstack)


class Network:
    def __init__(self, nodes: Sequence, edges: Sequence, name: str = "",
                 coordinates: np.ndarray = None):
        """
        Attributes:
            nodes (Sequence): data about nodes; it can also be just range(n)
            edges (List[Edges]): one or more set of edges
            name (str): network name

        Args:
            nodes (Sequence): data about nodes; it can also be just range(n)
            edges (sp.spmatrix or Edges or List[Edges]):
                one or more set of edges
            name (str): network name
        """
        def as_edges(edges):
            if isinstance(edges, Edges):
                return edges
            if sp.issparse(edges):
                return UndirectedEdges(edges)
            raise ValueError(
                "edges must be an instance of 'Edges' or a sparse matrix,"
                f"not '{type(edges).__name__}")

        self.nodes = nodes
        if isinstance(edges, Sequence):
            self.edges = [as_edges(e) for e in edges]
        else:
            self.edges = [as_edges(edges)]
        if not all(edges.edges.shape[0] == edges.edges.shape[1]
                   for edges in self.edges):
            raise ValueError("Edges must be square matrices")
        if not all(edges.edges.shape[0] == len(nodes) for edges in self.edges):
            raise ValueError("Number of nodes and edge matrix dimension must match")
        self.name = name
        self.coordinates = coordinates

    def copy(self):
        """Constructs a shallow copy of the network"""
        return type(self)(self.nodes, self.edges, self.name, self.coordinates)

    def number_of_nodes(self):
        return len(self.nodes)

    @sum_over_edge_types()
    def number_of_edges(self, edge_type=None):
        return len(self.edges[edge_type].edges.indices)

    def links(self, attr, edge_type=0, matrix_type=sp.coo_matrix):
        edges = self.edges[edge_type]
        return matrix_type(edges.edges), edges.edge_data.get_column(attr)

    @sum_over_edge_types()
    def out_degrees(self, edge_type, *, weighted=False):
        return self.edges[edge_type].out_degrees(weighted=weighted)

    @sum_over_edge_types()
    def in_degrees(self, edge_type=None, *, weighted=False):
        return self.edges[edge_type].in_degrees(weighted=weighted)

    @sum_over_edge_types()
    def degrees(self, edge_type=None, *, weighted=False):
        return self.edges[edge_type].degrees(weighted=weighted)

    @sum_over_edge_types(1)
    def out_degree(self, node, edge_type=None, *, weighted=False):
        return self.edges[edge_type].out_degree(node, weighted=weighted)

    @sum_over_edge_types(1)
    def in_degree(self, node, edge_type=None, *, weighted=False):
        return self.edges[edge_type].in_degree(node, weighted=weighted)

    @sum_over_edge_types(1)
    def degree(self, node, edge_type=None, *, weighted=False):
        return self.edges[edge_type].degree(node, weighted=weighted)

    @concatenate_over_edge_types(1)
    def outgoing(self, node, edge_type=None, weights=False):
        return self.edges[edge_type].outgoing(node, weights)

    @concatenate_over_edge_types(1)
    def incoming(self, node, edge_type=None, weights=False):
        return self.edges[edge_type].incoming(node, weights)

    @concatenate_over_edge_types(1)
    def neighbours(self, node, edge_type=None, weights=False):
        return self.edges[edge_type].neighbours(node, weights)

    def subgraph(self, mask):
        nodes = self.nodes[mask]
        if self.coordinates is not None:
            coordinates = self.coordinates[mask]
        else:
            coordinates = None
        if mask.dtype != bool:
            mask1 = np.full((self.number_of_nodes(),), False)
            mask1[mask] = True
            mask = mask1
        node_renumeration = np.cumsum(mask) - 1
        shape = (len(nodes), len(nodes))
        edge_sets = [edges.subset(mask, node_renumeration, shape)
                     for edges in self.edges]
        return Network(nodes, edge_sets, self.name, coordinates)



import numpy as np
from numpy.lib.stride_tricks import as_strided
import scipy.sparse as sp

from AnyQt.QtCore import Qt

from orangewidget.report import report
from Orange.data import Table
from Orange.distance import Euclidean
from Orange.misc import DistMatrix
from Orange.widgets import gui, widget, settings
from Orange.widgets.widget import Input, Output
from Orange.widgets.utils.widgetpreview import WidgetPreview

from orangecontrib.network.network import Network
from orangecontrib.network.network.base import UndirectedEdges, DirectedEdges
# This enables summarize in widget preview, pylint: disable=unused-import
import orangecontrib.network.widgets
from orangecontrib.network.widgets.utils import items_from_distmatrix, weights_from_distances


class OWNxNeighbor(widget.OWWidget):
    name = "Network of Neighbors"
    description = 'Constructs a network by connecting nearest neighbors.'
    icon = "icons/NetworkOfNeighbors.svg"
    priority = 6445

    class Inputs:
        distances = Input("Distances", DistMatrix)

    class Outputs:
        network = Output("Network", Network)

    resizing_enabled = False
    want_main_area = False

    k = settings.Setting(3)
    directed = settings.Setting(False)
    auto_apply = settings.Setting(True)

    def __init__(self):
        super().__init__()
        self.matrix = None
        self.symmetric = False
        self.graph = None

        box = gui.vBox(self.controlArea, box=True)
        gui.spin(
            box, self, 'k', 1, 20, label="Number of neighbors:",
            alignment=Qt.AlignRight, callback=self.commit.deferred,
            controlWidth=60)
        gui.checkBox(
            box, self, 'directed', 'Directed edges',
            callback=self.commit.deferred)

        gui.auto_apply(self.controlArea, self)

    @Inputs.distances
    def set_matrix(self, matrix: DistMatrix):
        if matrix is not None and matrix.size < 2:
            matrix = None
        if matrix is None:
            self.symmetric = False
            self.controls.k.setMaximum(1)
        else:
            self.controls.k.setMaximum(min(20, matrix.shape[0] - 1))
            self.symmetric = matrix.is_symmetric()
        self.matrix = matrix
        self.commit.now()

    @gui.deferred
    def commit(self):
        if self.matrix is None:
            self.graph = None
        else:
            edge_type = DirectedEdges if self.directed else UndirectedEdges
            items = items_from_distmatrix(self.matrix)
            edges = edge_type(self.get_neighbors().tocsr())
            self.graph = Network(items, edges)
        self.Outputs.network.send(self.graph)

    def get_neighbors(self):
        k = min(self.k, self.matrix.shape[0] - 1)
        matrix = np.asarray(self.matrix)
        if matrix.base is not None:
            matrix = matrix.copy()
        # TODO: use matrix.diag.fill(np.inf) when we require a newer numpy
        matrix[np.eye(len(matrix), dtype=bool)] = np.inf
        cols = np.argsort(matrix, axis=1)[:, :k].flatten().astype(np.int32)
        rows = np.repeat(np.arange(len(matrix), dtype=np.int32), k)
        if not self.directed:
            edges = np.array([rows, cols]).T
            edges.sort()
            cols, rows = np.unique(edges, axis=0).T

        weights = weights_from_distances(matrix[rows, cols])
        mask = weights > 0
        weights = weights[mask]

        return sp.coo_array((weights, (rows, cols)), matrix.shape)

    def send_report(self):
        self.report_items("Settings", [
            ("Neighbors", self.k),
            ("Directed", report.bool_str(self.directed))
        ])
        if self.graph is not None:
            self.report_items("Output network", [
                ("Vertices", self.graph.number_of_nodes()),
                ("Edges", self.graph.number_of_edges())])


if __name__ == "__main__":
    distances = Euclidean(Table("iris"))
    WidgetPreview(OWNxNeighbor).run(set_matrix=distances)



import unittest
import scipy.sparse as sp

from orangecontrib.network.network.base import Network, UndirectedEdges
from orangecontrib.network.widgets.OWNxAnalysis import density, avg_degree, OWNxAnalysis
from orangecontrib.network.widgets.tests.utils import NetworkTest, _create_net


class TestOWNxAnalysis(NetworkTest):
    def setUp(self):
        self.widget = self.create_widget(OWNxAnalysis)
        self.small_undir = _create_net(((0, 1, 1.0), (0, 2, 1.0), (1, 2, 1.0), (2, 3, 1.0)), n=5)
        self.small_dir = _create_net(((0, 1, 1.0), (0, 2, 1.0), (1, 2, 1.0), (2, 3, 1.0)), n=5, directed=True)
        self.empty_net = Network([], UndirectedEdges(sp.coo_matrix((0, 0))))

    def test_density(self):
        self.assertAlmostEqual(density(self.small_dir), 0.2)
        self.assertAlmostEqual(density(self.small_undir), 0.4)
        # n = 0 and n = 1 should not cause division problems
        self.assertAlmostEqual(density(self.empty_net), 0.0)
        self.assertAlmostEqual(density(self.empty_net), 0.0)

    def test_average_degree(self):
        self.assertAlmostEqual(avg_degree(self.small_dir), 0.8)
        self.assertAlmostEqual(avg_degree(self.small_undir), 1.6)
        self.assertAlmostEqual(avg_degree(self.empty_net), 0.0)

    def test_show_valid_indices_by_graph_type(self):
        """ Test that some statistics get disabled based on type of network on input """
        self.send_signal(self.widget.Inputs.network, self.small_undir)
        self.assertFalse(self.widget.method_cbs["number_strongly_connected_components"].isEnabled())
        self.assertFalse(self.widget.method_cbs["number_weakly_connected_components"].isEnabled())

        self.send_signal(self.widget.Inputs.network, self.small_dir)
        self.assertTrue(self.widget.method_cbs["number_strongly_connected_components"].isEnabled())
        self.assertTrue(self.widget.method_cbs["number_weakly_connected_components"].isEnabled())


if __name__ == '__main__':
    unittest.main()



from Orange.data import Table
from Orange.data.util import get_unique_names
from Orange.widgets import gui, widget, settings
from Orange.widgets.widget import Input, Output
from orangecontrib.network import Network
from orangecontrib.network.network import community as cd
from orangewidget.settings import rename_setting


class OWNxClustering(widget.OWWidget):
    name = 'Network Clustering'
    description = 'Orange widget for community detection in networks.'
    icon = "icons/NetworkClustering.svg"
    priority = 6430

    class Inputs:
        network = Input("Network", Network, default=True)

    class Outputs:
        network = Output("Network", Network)
        items = Output("Items", Table)

    resizing_enabled = False
    want_main_area = False

    settings_version = 2
    attenuate = settings.Setting(False)
    iterations = settings.Setting(1000)
    use_random_state = settings.Setting(False)
    hop_attenuation = settings.Setting(0.1)
    auto_apply = settings.Setting(True)

    def __init__(self):
        super().__init__()
        self.net = None
        self.cluster_feature = None

        box = gui.vBox(self.controlArea, "Label Propagation")
        gui.spin(
            box, self, "iterations", 1, 100000, 1,
            label="Max. iterations: ", callback=self.commit)
        gui.doubleSpin(box, self, "hop_attenuation", 0, 1, 0.01,
                       label="Apply hop attenuation: ",
                       checked="attenuate", callback=self.commit)
        self.random_state = gui.checkBox(
            box, self, "use_random_state",
            label="Replicable clustering", callback=self.commit)

        gui.auto_apply(self.controlArea, self)

    @Inputs.network
    def set_network(self, net):
        self.net = net
        self.commit()

    def commit(self):
        kwargs = {'iterations': self.iterations}
        if self.attenuate:
            alg = cd.label_propagation_hop_attenuation
            kwargs['delta'] = self.hop_attenuation
        else:
            alg = cd.label_propagation

        if self.net is None:
            self.Outputs.items.send(None)
            self.Outputs.network.send(None)
            self.cluster_feature = None
            return

        if self.use_random_state:
            kwargs['seed'] = 0

        labels = alg(self.net, **kwargs)
        domain = self.net.nodes.domain
        # Tie a name for presenting clustering results to the widget instance
        if self.cluster_feature is None:
            self.cluster_feature = get_unique_names(domain, 'Cluster')

        net = self.net.copy()
        cd.add_results_to_items(net, labels, self.cluster_feature)

        self.Outputs.items.send(net.nodes)
        self.Outputs.network.send(net)

        nclusters = len(set(labels.values()))

    @classmethod
    def migrate_settings(cls, settings, version):
        if version < 2:
            rename_setting(settings, "autoApply", "auto_apply")
            if hasattr(settings, "method"):
                rename_setting(settings, "method", "attenuate")


if __name__ == "__main__":
    from Orange.widgets.utils.widgetpreview import WidgetPreview
    from orangecontrib.network.network.readwrite \
        import read_pajek, transform_data_to_orange_table
    from os.path import join, dirname

    fname = join(dirname(dirname(__file__)), 'networks', 'leu_by_genesets.net')
    network = read_pajek(fname)
    transform_data_to_orange_table(network)
    WidgetPreview(OWNxClustering).run(set_network=network)
